import pandas as pd
from functools import partial
from circmimi.reference.utils import cwd, open_file
from circmimi.reference.resource import MiRDBResource, Gene2AccessionResource
from circmimi.reference.species import species_list


def get_needed_data(filename, data_filter):
    with open_file(filename) as f_in:
        for line in data_filter(f_in):
            yield line.rstrip('\n').split('\t')


def mirdb_filter(data, species_key):
    for line in data:
        if line.startswith(species_key):
            yield line


def gene_accession_filter(data, species_tax_id):
    species_tax_id_txt = "{}\t".format(species_tax_id)

    title = data.readline()
    yield title

    for line in data:
        if line.startswith(species_tax_id_txt):
            yield line


def generate(species, version, ref_dir, out_file, show_accession=False):
    species = species_list[species]

    with cwd(ref_dir):
        mirdb_file = MiRDBResource(None, version)
        gene2accession = Gene2AccessionResource()

        # download
        mirdb_file.download()
        gene2accession.download()

        # basic filter for the "species"
        ## mirdb
        mirdb_data = get_needed_data(
            mirdb_file.filename,
            data_filter=partial(mirdb_filter, species_key=species.key)
        )

        mirdb_df = pd.DataFrame(
            mirdb_data,
            columns=['mirna', 'accession', 'targeting_score']
        )

        ## gene2accesion
        gene_accession_data = get_needed_data(
            gene2accession.filename,
            data_filter=partial(
                gene_accession_filter,
                species_tax_id=species.tax_id
            )
        )

        gene_accession_data_title = next(gene_accession_data)

        gene_accession_df = pd.DataFrame(
            gene_accession_data,
            columns=gene_accession_data_title
        ).pipe(
            lambda df: df[['RNA_nucleotide_accession.version', 'Symbol', 'GeneID']]
        ).drop_duplicates(
        ).pipe(
            lambda df: df[df['RNA_nucleotide_accession.version'] != '-']
        )

        accession_no_version_df = gene_accession_df.apply(
            lambda s: pd.Series(s['RNA_nucleotide_accession.version'].split('.')[0]),
            axis=1
        )

        gene_accession_df_2 = gene_accession_df.assign(
            accession=accession_no_version_df
        ).pipe(
            lambda df: df[['accession', 'Symbol']]
        ).rename(
            {
                'Symbol': 'target_gene'
            },
            axis=1
        )

        # merge
        mirdb_df_with_symbol = mirdb_df.merge(
            gene_accession_df_2,
            on='accession',
            how='inner'
        ).pipe(
            lambda df: df[['mirna', 'target_gene', 'targeting_score', 'accession']]
        )

        if not show_accession:
            mirdb_df_with_symbol = mirdb_df_with_symbol.drop(
                'accession',
                axis=1
            ).groupby(
                [
                    'mirna',
                    'target_gene'
                ]
            ).agg('max').reset_index()

        mirdb_df_with_symbol.to_csv(out_file, sep='\t', index=None)



import io
import pandas as pd
import subprocess as sp
import tempfile as tp
from functools import partial
from collections import deque


class Blat:
    def __init__(self, work_dir='.', blat_bin='blat', blat_opts="", num_proc=1, mp_blat_bin='mp_blat.py'):
        self.work_dir = work_dir
        self.blat_bin = blat_bin
        self.blat_opts = blat_opts
        self.num_proc = num_proc
        self.mp_blat_bin = mp_blat_bin

    def _run(self, ref_file, fa_file):
        with tp.NamedTemporaryFile(dir=self.work_dir) as tmp_out:
            cmd = self._generate_cmd(ref_file, fa_file, tmp_out.name)
            sp.run(cmd)

            result = BlatPsl(tmp_out.read().decode('utf-8'))
            return result

    def _generate_cmd(self, ref_file, fa_file, out_file):
        cmd = [self.mp_blat_bin, ref_file, fa_file, out_file]
        cmd += ['-p', str(self.num_proc)]
        cmd += ['--tmp_path', self.work_dir]
        cmd += ['--blat_bin', self.blat_bin]
        cmd += ['--blat_opt', self.blat_opts]
        return cmd

    def __call__(self, ref_file, fa_file):
        return self._run(ref_file, fa_file)


class BlatPsl:
    PSL_TITLE = (
        'matches',
        'misMatches',
        'repMatches',
        'nCount',
        'qNumInsert',
        'qBaseInsert',
        'tNumInsert',
        'tBaseInsert',
        'strand',
        'qName',
        'qSize',
        'qStart',
        'qEnd',
        'tName',
        'tSize',
        'tStart',
        'tEnd',
        'blockCount',
        'blockSizes',
        'qStarts',
        'tStarts'
    )

    def __init__(self, psl_text):
        self.psl_text = psl_text
        self.df = self.parse(psl_text)

    @classmethod
    def parse(cls, psl_text):
        no_header_psl = cls.remove_header(psl_text)

        with io.StringIO() as tmp_out:
            tmp_out.write(no_header_psl)
            tmp_out.seek(0)

            psl_df = pd.read_csv(
                tmp_out,
                sep='\t',
                header=None,
                names=cls.PSL_TITLE
            )

            return psl_df

    @staticmethod
    def remove_header(psl_text):
        if psl_text.startswith('psLayout'):
            no_header_psl = psl_text.split('\n', 5)[5]
        else:
            no_header_psl = psl_text

        return no_header_psl


class PslFilters:
    @staticmethod
    def _is_colinear(df, threshold=0.8):
        return ((df.matches + df.repMatches) / df.qSize) > threshold

    @classmethod
    def colinear_filter(cls, df, threshold=0.8):
        _is_colinear = partial(cls._is_colinear, threshold=threshold)
        return df[_is_colinear]

    @staticmethod
    def _is_chimera(subdf, threshold):
        chimera = PslChimera()
        chimera.parse(subdf)

        if not chimera.linear:
            if (threshold.check(chimera.five_end) is True) and \
               (threshold.check(chimera.three_end) is True):
                return True

        return False

    @classmethod
    def chimera_filter(cls, df, min_matches=30, min_diff=3):
        threshold = _Threshold(min_matches=min_matches, min_diff=min_diff)
        _is_chimera = partial(cls._is_chimera, threshold=threshold)
        chimera_df = df.groupby('qName').filter(_is_chimera)
        return chimera_df


class _Threshold:
    def __init__(self, min_matches, min_diff):
        self.min_matches = min_matches
        self.min_diff = min_diff

    def check(self, matches):
        num_matches = len(matches)
        if num_matches == 0:
            best_match_score = 0
            second_best_score = 0
        elif num_matches == 1:
            best_match_score = matches[0].matches
            second_best_score = 0
        elif num_matches == 2:
            best_match_score = matches[0].matches
            second_best_score = matches[1].matches

        if best_match_score < self.min_matches:
            return False

        if best_match_score - second_best_score < self.min_diff:
            return False

        return True


class PslChimera:
    def __init__(self):
        self._linear = False
        self._five_end = deque(maxlen=2)
        self._three_end = deque(maxlen=2)

    def parse(self, df):
        for idx, match in df.iterrows():
            if self._is_colinear(match):
                self._linear = True

            if self._is_five_end(match):
                self._add_best_match(self._five_end, match)

            if self._is_three_end(match):
                self._add_best_match(self._three_end, match)

    @property
    def linear(self):
        return self._linear

    @property
    def five_end(self):
        return self._five_end

    @property
    def three_end(self):
        return self._three_end

    @staticmethod
    def _is_colinear(match):
        return (match.qStart < 10) and (match.qSize - match.qEnd < 10)

    @staticmethod
    def _is_five_end(match):
        return match.qStart < 10

    @staticmethod
    def _is_three_end(match):
        return match.qSize - match.qEnd < 10

    @staticmethod
    def _is_better_match(match, other):
        return match.matches >= other.matches

    @classmethod
    def _add_best_match(cls, q, match):
        len_q = len(q)

        if len_q == 0:
            q.append(match)
        elif len_q == 1:
            if cls._is_better_match(match, q[0]):
                q.appendleft(match)
            else:
                q.append(match)
        else:
            if cls._is_better_match(match, q[0]):
                q.appendleft(match)
            elif cls._is_better_match(match, q[1]):
                q.appendleft(match)
                q.reverse()
            else:
                pass


class PslUtils:
    @staticmethod
    def get_uniq_qname(df):
        return df.qName.unique()

    @staticmethod
    def remove_in_list(df, qnames):
        return df[~df.qName.isin(qnames)]



import pandas as pd
import numpy as np
import tempfile as tp
from functools import reduce
from circmimi.bed import BedUtils
from circmimi.seq import Seq
from circmimi.blat import Blat, PslFilters, PslUtils
from circmimi.annotation import Annotation


class AmbiguousChecker:
    _CHECK_LIST = [
        'ambiguity with an co-linear explanation',
        'ambiguity with multiple hits'
    ]

    def __init__(self,
                 anno_db_file,
                 ref_file,
                 other_ref_file,
                 work_dir='.',
                 num_proc=1,
                 blat_bin='blat',
                 mp_blat_bin='mp_blat.py'):
        self.work_dir = work_dir
        self.num_proc = num_proc

        self.blat_bin = blat_bin
        self.mp_blat_bin = mp_blat_bin

        self.anno_db = Annotation(anno_db_file)
        self.ref_file = ref_file
        self.other_ref_file = other_ref_file

    def check(self, circ_df):
        self._init_status(circ_df)

        # 1. get flanking sequences
        flanking_regions = circ_df.apply(
            self._get_flanking_region,
            axis=1
        )
        flanking_regions_df = self._to_regions_df(flanking_regions).dropna()
        flanking_seq_df = flanking_regions_df.pipe(
            BedUtils.to_bed_df
        ).pipe(
            Seq.get_seq,
            ref_file=self.ref_file
        )

        fa_file = tp.NamedTemporaryFile(dir=self.work_dir)
        fa_file.write(Seq.to_fasta(flanking_seq_df).encode('utf-8'))
        fa_file.seek(0)

        # 2. blat
        blat = Blat(
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            blat_bin=self.blat_bin,
            mp_blat_bin=self.mp_blat_bin
        )
        web_blat = Blat(
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            blat_opts="-tileSize=9 -stepSize=9 -repMatch=32768",
            blat_bin=self.blat_bin,
            mp_blat_bin=self.mp_blat_bin
        )

        self._psl_rG_1 = blat(self.ref_file, fa_file.name)
        self._psl_rG_2 = web_blat(self.ref_file, fa_file.name)
        self._psl_rO_1 = blat(self.other_ref_file, fa_file.name)
        self._psl_rO_2 = web_blat(self.other_ref_file, fa_file.name)

        fa_file.close()

        # 3. evaluate
        # 3.1 colinear part
        psl_rG_1_colinear_df = self._psl_rG_1.df.pipe(PslFilters.colinear_filter)
        psl_rG_2_colinear_df = self._psl_rG_2.df.pipe(PslFilters.colinear_filter)
        psl_rO_1_colinear_df = self._psl_rO_1.df.pipe(PslFilters.colinear_filter)
        psl_rO_2_colinear_df = self._psl_rO_2.df.pipe(PslFilters.colinear_filter)

        psl_rG_1_colinear_ids = PslUtils.get_uniq_qname(psl_rG_1_colinear_df)
        psl_rG_2_colinear_ids = PslUtils.get_uniq_qname(psl_rG_2_colinear_df)
        psl_rO_1_colinear_ids = PslUtils.get_uniq_qname(psl_rO_1_colinear_df)
        psl_rO_2_colinear_ids = PslUtils.get_uniq_qname(psl_rO_2_colinear_df)

        self._all_colinear_ids = reduce(
            np.union1d,
            [
                psl_rG_1_colinear_ids,
                psl_rG_2_colinear_ids,
                psl_rO_1_colinear_ids,
                psl_rO_2_colinear_ids
            ]
        )

        # 3.2 multiple hits
        psl_rG_1_chimera_df = self._psl_rG_1.df.pipe(PslFilters.chimera_filter)
        psl_rG_2_chimera_df = self._psl_rG_2.df.pipe(PslFilters.chimera_filter)

        self._psl_rG_1_chimera_ids = PslUtils.get_uniq_qname(psl_rG_1_chimera_df)
        self._psl_rG_2_chimera_ids = PslUtils.get_uniq_qname(psl_rG_2_chimera_df)

        psl_rG_1_multiple_hits_df = self._psl_rG_1.df.pipe(
            PslUtils.remove_in_list,
            qnames=np.union1d(
                psl_rG_1_colinear_ids,
                self._psl_rG_1_chimera_ids
            )
        )

        psl_rG_2_multiple_hits_df = self._psl_rG_2.df.pipe(
            PslUtils.remove_in_list,
            qnames=np.union1d(
                psl_rG_2_colinear_ids,
                self._psl_rG_2_chimera_ids
            )
        )

        psl_rG_1_multiple_hits_ids = PslUtils.get_uniq_qname(psl_rG_1_multiple_hits_df)
        psl_rG_2_multiple_hits_ids = PslUtils.get_uniq_qname(psl_rG_2_multiple_hits_df)

        self._all_multiple_hits_ids = np.setdiff1d(
            np.union1d(
                psl_rG_1_multiple_hits_ids,
                psl_rG_2_multiple_hits_ids
            ),
            self._all_colinear_ids
        )

        for id_ in self._all_colinear_ids:
            self._report_status(id_, self._CHECK_LIST[0])

        for id_ in self._all_multiple_hits_ids:
            self._report_status(id_, self._CHECK_LIST[1])

        return self._checking_result

    def _get_flanking_region(self, circ_data):
        chr_ = circ_data.chr
        donor_site = circ_data.donor
        acceptor_site = circ_data.acceptor
        strand = circ_data.strand

        ev_id = circ_data.name

        donor = self.anno_db\
            .get_nearest_donor_site(chr_, donor_site, strand, dist=5)
        acceptor = self.anno_db\
            .get_nearest_acceptor_site(chr_, acceptor_site, strand, dist=5)

        if (donor is None) or (acceptor is None):
            self._report_status(ev_id, self._CHECK_LIST[0], np.nan)
            self._report_status(ev_id, self._CHECK_LIST[1], np.nan)
            return np.nan

        donor_site = donor.junc_site
        donor_exon = max(donor.exons, key=lambda exon: len(exon))
        donor_acceptor = donor_exon.acceptor.junc_site

        acceptor_site = acceptor.junc_site
        acceptor_exon = max(acceptor.exons, key=lambda exon: len(exon))
        acceptor_donor = acceptor_exon.donor.junc_site

        if strand == '+':

            donor_flanking = [
                chr_,
                max(donor_site - 99, donor_acceptor),
                donor_site,
                strand
            ]

            acceptor_flanking = [
                chr_,
                acceptor_site,
                min(acceptor_site + 99, acceptor_donor),
                strand
            ]

        elif strand == "-":
            donor_flanking = [
                chr_,
                donor_site,
                min(donor_site + 99, donor_acceptor),
                strand
            ]

            acceptor_flanking = [
                chr_,
                max(acceptor_site - 99, acceptor_donor),
                acceptor_site,
                strand
            ]

        return [donor_flanking, acceptor_flanking]

    @staticmethod
    def _to_regions_df(regions):
        df = pd.DataFrame(
            regions,
            columns=['regions']
        ).rename_axis('regions_id').reset_index()

        return df

    def _init_status(self, circ_df):
        self._checking_result = pd.DataFrame(
            [],
            columns=self._CHECK_LIST
        ).rename_axis('ev_id')

        for id_ in circ_df.index:
            self._report_status(id_)

    def _report_status(self, ev_id, status=None, value='1', init_value='0'):
        if ev_id in self._checking_result.index:
            if status is not None:
                self._checking_result.loc[ev_id, status] = value
        else:
            ev_status = pd.Series(
                [init_value] * len(self._CHECK_LIST),
                index=self._CHECK_LIST,
                name=ev_id
            )

            if status is not None:
                ev_status[status] = value

            self._checking_result = self._checking_result.append(ev_status)



import networkx as nx
from circmimi.network.utils import read_table
from circmimi.network.formats.xgmml import write_xgmml


class CyNetwork:
    def __init__(self):
        self.graph = nx.DiGraph()

    def load_data(self, file_):
        for data in read_table(file_):
            circRNA_id = ':'.join(data[:5])

            self.graph.add_node(
                circRNA_id,
                data_type="circRNA",
                data=dict(list(data._asdict().items())[:5])
            )

            self.graph.add_node(
                data.mirna,
                data_type="miRNA",
                data={'name': data.mirna}
            )

            self.graph.add_node(
                data.target_gene,
                data_type='target_gene',
                data={'name': data.target_gene}
            )

            self.graph.add_edge(
                circRNA_id,
                data.mirna,
                data_type="circ-mir",
                data=dict(list(data._asdict().items())[6:9])
            )

            self.graph.add_edge(
                data.mirna,
                data.target_gene,
                data_type="mir-target",
                data=dict(list(data._asdict().items())[10:])
            )

    def update_node(self, node, key_val):
        self.graph.nodes[node].update(key_val)

    def update_edge(self, edge, key_val):
        self.graph.edges[edge].update(key_val)

    def apply_layout(self, layout):
        pass

    def apply_style(self, style):
        pass

    def to_xgmml(self, file_):
        write_xgmml(self.graph, file_)


class Layout:
    def __init__(self):
        pass

    def apply(self, graph):
        return graph


class Style:
    def __init__(self):
        pass

    def apply(self, graph):
        return graph



import re
from operator import itemgetter
from circmimi.reference import resource as rs
from circmimi.reference.utils import open_file


class MatureMiRNAUpdater:
    def __init__(self, from_, to_, species):
        self.from_ = from_
        self.to_ = to_
        self.species = species
        self.mapping_table = {}

    def create(self, ref_dir='.'):
        dat_file_from_ = rs.MiRBaseDat(self.species, self.from_)
        dat_file_to_ = rs.MiRBaseDat(self.species, self.to_)
        diff_file = rs.MiRBaseDiff(self.species, self.to_)

        dat_file_from_.download(ref_dir)
        dat_file_from_.rename_with_version()
        dat_file_to_.download(ref_dir)
        dat_file_to_.rename_with_version()
        diff_file.download(ref_dir)

        dat_text_from_ = self._load_data(dat_file_from_.filename)
        dat_text_to_ = self._load_data(dat_file_to_.filename)
        diff_text = self._load_data(diff_file.filename)

        mature_diff_data = self._get_mature_diff(diff_text, self.species)

        renamed_accessions = self._get_accessions_of_type('NAME', mature_diff_data)
        renamed_miRNAs_from_ = self._get_IDs_by_accessions(dat_text_from_, renamed_accessions)
        renamed_miRNAs_to_ = self._get_IDs_by_accessions(dat_text_to_, renamed_accessions)
        self.mapping_table.update(dict(zip(renamed_miRNAs_from_, renamed_miRNAs_to_)))

        deleted_accessions = self._get_accessions_of_type('DELETE', mature_diff_data)
        deleted_miRNAs = list(self._get_IDs_by_accessions(dat_text_from_, deleted_accessions))
        self.mapping_table.update({mirna_id: '' for mirna_id in deleted_miRNAs})

    @staticmethod
    def _get_mature_diff(diff_text, species):
        mature_diff_text = re.search(
            r'#\n# Mature sequences start here\n#\n(.*)',
            diff_text,
            flags=re.DOTALL
        ).group(1)

        diff_of_species = re.findall(
            r'(.*)\t({}-.*)\t(.*)'.format(species),
            mature_diff_text
        )

        return diff_of_species

    @staticmethod
    def _get_accessions_of_type(diff_type, diff_data):
        diff_data_of_type = filter(lambda data: data[2] == diff_type, diff_data)
        accessions = list(map(itemgetter(0), diff_data_of_type))
        return accessions

    @staticmethod
    def _get_ID_by_accession(mirna_dat, accession):
        m = re.search(r'accession="{}"\n.*product="([^"]*)"'.format(accession), mirna_dat)
        if m:
            return m.group(1)
        else:
            return ''

    @classmethod
    def _get_IDs_by_accessions(cls, mirna_dat, accessions):
        for accession in accessions:
            yield cls._get_ID_by_accession(mirna_dat, accession)

    @staticmethod
    def _load_data(filename):
        with open_file(filename) as f_in:
            data_text = f_in.read()

        return data_text

    def save(self, mapping_file):
        with open(mapping_file, 'w') as out:
            for k, v in self.mapping_table.items():
                print(k, v, sep='\t', file=out)

    def load_maps(self, mapping_file):
        with open(mapping_file) as f_in:
            for line in f_in:
                data = line.rstrip('\n').split('\t')
                self.mapping_table[data[0]] = data[1]

    def update(self, mirna):
        return self.mapping_table.get(mirna, mirna)

    def update_row(self, row_data, col_key=0, inplace=False, remove_deleted=False):
        updated_row = list(row_data)
        updated_value = self.update(updated_row[col_key])

        if remove_deleted:
            if updated_value == '':
                return

        if inplace:
            updated_row[col_key] = updated_value
        else:
            updated_row.append(updated_value)

        return updated_row

    def update_file(self, in_file, out_file, col_key=0, inplace=False, remove_deleted=False):
        with open(in_file) as f_in, open(out_file, 'w') as f_out:
            for line in f_in:
                data = line.rstrip('\n').split('\t')
                updated_data = self.update_row(data, col_key, inplace, remove_deleted)
                if updated_data is not None:
                    print(*updated_data, sep='\t', file=f_out)



from collections import namedtuple


def read_table(filename, sep='\t', titles=None):
    """A generator function for reading data.

    Args:
      filename: the path to the data file

    Yield:
      Data: a namedtuple with titles from the data
    """

    with open(filename) as f_in:

        if titles is None:
            titles = f_in.readline().rstrip('\n').split(sep)

        Data = namedtuple('Data', titles)

        for line in f_in:
            data = Data(*line.rstrip('\n').split(sep))
            yield data



'''
Use case:

from .resource import (EnsemblGenome,
                               EnsemblAnnotation,
                               GencodeGenome,
                               GencodeAnnotation)

reference_genome = EnsemblGenome('homo_sapiens', '97')
reference_genome.download('./refs')
genome_filename = reference_genome.filename

'''


import subprocess as sp
import ftplib
import os
import re


class Resource:
    def __init__(self, url):
        self.url = url
        self.filename = os.path.basename(url)

    def download(self, dir_='.'):
        dest_path = os.path.join(dir_, self.filename)

        if os.path.exists(dest_path):
            self.filename = dest_path

        elif os.path.exists(dest_path.rstrip('.gz')):
            self.filename = dest_path.rstrip('.gz')

        else:
            res = sp.run(['wget', self.url, '-P', dir_])

            if res.returncode == 0:
                self.filename = dest_path

        return self.filename

    def unzip(self):
        if self.filename.endswith('.gz'):
            res = sp.run(['gunzip', self.filename])

            if res.returncode == 0:
                self.filename = re.sub(r'\.gz$', '', self.filename)

        return self.filename

    def rename(self, dest_name):
        if os.path.exists(dest_name):
            raise FileExistsError

        os.renames(self.filename, dest_name)
        self.filename = dest_name

        return self.filename


class FTPResource(Resource):
    ftp_site = ""
    release_dir = ""
    release_pattern = r""
    file_pattern = r""
    another_pats = ()

    def __init__(self, species, version):
        self.species = species
        self.version = version
        url = self.get_url()
        super().__init__(url)

    def get_url(self):
        with ftplib.FTP(self.ftp_site) as ftp:
            ftp.login()

            if self.version == "current":
                self.version = self.get_current_release(ftp)

            dir_path = self.get_dir_path()

            try:
                ftp.cwd(dir_path)
            except ftplib.error_perm:
                raise URLError(dir_path) from None

            dir_list = ftp.nlst()

            filename = self.get_the_filename(dir_list)

            url = "ftp://{}{}".format(
                self.ftp_site,
                os.path.join(dir_path, filename)
            )

            return url

    def get_current_release(self, ftp, max_key=float):
        ftp.cwd(self.release_dir.format(species=self.species))
        dir_list = ftp.nlst()
        current_release = max(map(self.get_release_number, dir_list), key=max_key)
        return current_release

    @classmethod
    def get_release_number(cls, name):
        m = re.search(cls.release_pattern, name)
        if m:
            return m.group(1)
        else:
            return '-1'

    def get_dir_path(self):
        raise NotImplementedError

    @classmethod
    def get_the_filename(cls, dir_list):
        for name in dir_list:
            m1 = re.search(cls.file_pattern, name)
            if m1:
                return m1.group(0)
        else:
            for pat in cls.another_pats:
                for name in dir_list:
                    m2 = re.search(pat, name)
                    if m2:
                        return m2.group(0)
            else:
                return ""


class EnsemblResource(FTPResource):
    ftp_site = "ftp.ensembl.org"
    release_dir = "/pub"
    release_pattern = r"^release-([0-9]+)$"


class EnsemblGenome(EnsemblResource):
    file_pattern = r".+\.dna\.primary_assembly\.fa\.gz$"
    another_pats = (r".+\.dna\.toplevel\.fa\.gz$",)

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            'fasta',
            self.species,
            'dna'
        )

        return dir_path


class EnsemblCDna(EnsemblResource):
    file_pattern = r".+\.cdna\.all\.fa\.gz$"

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            'fasta',
            self.species,
            'cdna'
        )

        return dir_path


class EnsemblNCRna(EnsemblResource):
    file_pattern = r".+\.ncrna\.fa\.gz$"

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            'fasta',
            self.species,
            'ncrna'
        )

        return dir_path


class EnsemblAnnotation(EnsemblResource):
    file_pattern = r".+\.chr\.gtf\.gz$"
    another_pats = (r".+\.[0-9]+\.gtf\.gz$",)

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            'gtf',
            self.species
        )

        return dir_path


class EnsemblSisterResource(FTPResource):
    ftp_site = "ftp.ensemblgenomes.org"
    release_dir = "/pub"
    release_pattern = r"^release-([0-9]+)$"

    def __init__(self, field, species, version):
        self.field = field
        super().__init__(species, version)


class EnsemblSisterGenome(EnsemblSisterResource):
    file_pattern = r".+\.dna\.primary_assembly\.fa\.gz$"
    another_pats = (r".+\.dna\.toplevel\.fa\.gz$",)

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            self.field,
            'fasta',
            self.species,
            'dna'
        )

        return dir_path


class EnsemblSisterCDna(EnsemblSisterResource):
    file_pattern = r".+\.cdna\.all\.fa\.gz$"

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            self.field,
            'fasta',
            self.species,
            'cdna'
        )

        return dir_path


class EnsemblSisterNCRna(EnsemblSisterResource):
    file_pattern = r".+\.ncrna\.fa\.gz$"

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            self.field,
            'fasta',
            self.species,
            'ncrna'
        )

        return dir_path


class EnsemblSisterAnnotation(EnsemblSisterResource):
    file_pattern = r".+\.chr\.gtf\.gz$"
    another_pats = (r".+\.[0-9]+\.gtf\.gz$",)

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub',
            'release-{}'.format(self.version),
            self.field,
            'gtf',
            self.species
        )

        return dir_path


class GencodeResource(FTPResource):
    ftp_site = "ftp.ebi.ac.uk"
    release_dir = "/pub/databases/gencode/Gencode_{species}/"
    release_pattern = r"^release_(M?[0-9]+)$"

    def get_current_release(self, ftp, max_key=float):
        if self.species == "mouse":
            max_key = self.get_digit_part

        return super().get_current_release(ftp, max_key)

    @staticmethod
    def get_digit_part(version):
        m = re.search(r'[^0-9]*([0-9]+)', version)
        if m:
            digit = float(m.group(1))
        else:
            digit = -1

        return digit

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub/databases/gencode',
            'Gencode_{}'.format(self.species),
            'release_{}'.format(self.version)
        )

        return dir_path


class GencodeGenome(GencodeResource):
    file_pattern = r".+\.primary_assembly\.genome\.fa\.gz$"


class GencodeProteinCodingTranscripts(GencodeResource):
    file_pattern = r".+\.lncRNA_transcripts\.fa\.gz$"


class GencodeLongNonCodingTranscripts(GencodeResource):
    file_pattern = r".+\.pc_transcripts\.fa\.gz$"


class GencodeAnnotation(GencodeResource):
    file_pattern = r"^gencode.v[^.]+.annotation.gtf.gz$"


class MiRBaseResource(FTPResource):
    ftp_site = "mirbase.org"
    release_dir = "/pub/mirbase/"
    release_pattern = r"([0-9]+(?:\.[0-9]+)?)"

    def get_dir_path(self):
        dir_path = os.path.join(
            '/',
            'pub/mirbase',
            '{}'.format(self.version)
        )

        return dir_path

    def rename_with_version(self):
        dirname, filename = os.path.split(self.filename)
        name_with_ver = self._add_version_to_name(filename, self.version)
        dest_name = os.path.join(dirname, name_with_ver)
        self.rename(dest_name)

        return self.filename

    @staticmethod
    def _add_version_to_name(filename, version):
        sep_name = filename.split('.')
        sep_name.insert(1, "v{}".format(version))
        return '.'.join(sep_name)


class MiRBaseMiRNA(MiRBaseResource):
    file_pattern = r"mature\.fa\.gz"


class MiRBaseDiff(MiRBaseResource):
    file_pattern = r"miRNA\.diff\.gz"


class MiRBaseDat(MiRBaseResource):
    file_pattern = r"miRNA\.dat\.gz"


class MiRTarBaseResource(Resource):
    url_templ = "http://mirtarbase.mbc.nctu.edu.tw/cache/download/{version}/miRTarBase_MTI.xlsx"

    def __init__(self, species, version):
        self.species = species
        self.version = version
        url = self.get_url()
        super().__init__(url)

    def get_url(self):
        url = self.url_templ.format(species=self.species, version=self.version)
        return url


class MiRDBResource(Resource):
    url_templ = "http://www.mirdb.org/download/miRDB_v{version}_prediction_result.txt.gz"
    url_templ_2 = "http://www.mirdb.org/download/MirTarget2_v{version}_prediction_result.txt.gz"

    def __init__(self, species, version):
        self.species = species
        self.version = version
        url = self.get_url()
        super().__init__(url)

    def get_url(self):
        if self.version in ['1.0', '2.0', '3.0', '4.0']:
            url_templ = self.url_templ_2
        else:
            url_templ = self.url_templ

        url = url_templ.format(
            species=self.species,
            version=self.version
        )

        return url


class Gene2AccessionResource(Resource):
    url = "ftp://ftp.ncbi.nih.gov/gene/DATA/gene2accession.gz"

    def __init__(self):
        super().__init__(self.url)


class MiRDBData(Resource):
    url_templ = "ftp://treeslab1.genomics.sinica.edu.tw/CircMiMi/miRDB_data/v{version}/miRDB_v{version}_{species}.tsv.gz"
    available_species = ('hsa', 'mmu', 'rno', 'cfa', 'gga')

    def __init__(self, species, version):

        if species in self.available_species:
            self.species = species
        else:
            self.species = "others"

        self.version = version
        url = self.get_url()
        super().__init__(url)

    def get_url(self):
        url = self.url_templ.format(species=self.species, version=self.version)
        return url


class Error(Exception):
    pass

class URLError(Error):
    pass

class SourceNotSupportError(Error):
    pass



import pandas as pd
import io


class Bed:
    BED_TITLE = (
        'chr',
        'start',
        'end',
        'name',
        'score',
        'strand',
        'thickStart',
        'thickEnd',
        'itemRGB',
        'blockCount',
        'blockSizes',
        'blockStarts'
    )

    def __init__(self, regions, name='.'):
        self.regions = regions
        self.name = name

        self._parse_regions(regions)

    def _parse_regions(self, regions):
        self._parse_region(regions[0])
        for region in regions[1:]:
            self._add_block(region)

        if self.strand == '-':
            self._reverse_region()

    def _parse_region(self, region):
        self.chrom = region[0]
        self.start = int(region[1]) - 1
        self.end = int(region[2])
        self.strand = region[3]

        self.block_count = 1
        self.block_sizes = [self.end - self.start]
        self.block_starts = [0]

    def _add_block(self, region):
        other = Bed([region])

        assert self.strand == other.strand

        self.block_count += other.block_count
        self.block_sizes += other.block_sizes

        all_starts = [
            start + self.start for start in self.block_starts
        ] + [other.start]

        if other.start < self.start:
            self.start = other.start

        self.block_starts = [start - self.start for start in all_starts]

        if other.end > self.end:
            self.end = other.end

    def _reverse_region(self):
        self.block_sizes = list(reversed(self.block_sizes))
        self.block_starts = list(reversed(self.block_starts))

    def get_data(self, all_fields=False):
        fields = [
            self.chrom,
            self.start,
            self.end,
            self.name,
            '.',
            self.strand
        ]

        if all_fields:
            fields += [
                self.start,
                self.end,
                0,
                self.block_count,
                self._list_to_str(self.block_sizes, sep=','),
                self._list_to_str(self.block_starts, sep=',')
            ]

        return fields

    def to_string(self, all_fields=False):
        data = self.get_data(all_fields=all_fields)
        bed_txt = self._list_to_str(data)
        return bed_txt

    @staticmethod
    def _list_to_str(list_, sep='\t', end=''):
        with io.StringIO() as tmp:
            print(*list_, sep=sep, end=end, file=tmp)
            return tmp.getvalue()


class BedUtils:
    @classmethod
    def to_regions_df(cls, exons_df):
        if exons_df.empty:
            regions_df = pd.DataFrame([], columns=['regions_id', 'regions'])
        else:
            regions_df = exons_df[['exons_id', 'exons']].assign(
                regions_id=lambda df: df.exons_id,
                regions=lambda df: df.exons.apply(cls._exons_to_regions)
            )[['regions_id', 'regions']]

        return regions_df

    @staticmethod
    def _exons_to_regions(exons):
        regions = list(map(lambda exon: exon.region, exons))
        return regions

    @classmethod
    def to_bed_df(cls, regions_df):
        if regions_df.empty:
            bed_df = pd.DataFrame([], columns=Bed.BED_TITLE)
        else:
            bed_df = regions_df.apply(cls._to_bed, axis=1)
            bed_df.columns = Bed.BED_TITLE

        return bed_df

    @classmethod
    def _to_bed(cls, s):
        bed = Bed(s.regions, name=s.regions_id)
        return pd.Series(bed.get_data(all_fields=True))



import re
import os
import pandas as pd
import csv
import shutil
from circmimi.reference import gendb
from circmimi.reference.species import species_list
from circmimi.seq import parse_fasta
from circmimi.reference import resource as rs
from circmimi.reference.utils import cwd
from circmimi.reference.mirbase import MatureMiRNAUpdater


class RefFile:
    def __init__(self, src_name):
        self.src_name = src_name
        self.filename = None

    def generate(self):
        return self.filename


class AnnoRef(RefFile):
    def generate(self):
        self.filename = re.sub(r'\.gtf(?:\.gz)?$', '.db', self.src_name)
        gendb.generate(self.src_name, self.filename)
        return self.filename


class MirRef(RefFile):
    def generate(self, species_key):
        self.filename = re.sub(r'\.fa$', '.{}.fa'.format(species_key), self.src_name)

        with open(self.src_name) as fa_in:
            fa_txt = fa_in.read()

        with open(self.filename, 'w') as out:
            for fa_id, fa_seq in parse_fasta(fa_txt):
                m = re.search(r'^{}-'.format(species_key), fa_id)
                if m:
                    print(">{}\n{}".format(fa_id, fa_seq), file=out)

        return self.filename


class MiRTarBaseRef(RefFile):
    def generate(self, species):
        df = pd.read_excel(self.src_name)

        formatted_data = df[
            df['Species (miRNA)'] == species.fullname
        ][[
            'miRNA',
            'Target Gene',
            'References (PMID)'
        ]].groupby([
            'miRNA',
            'Target Gene'
        ]).agg(
            'count'
        ).reset_index(
        ).rename(
            {
                'miRNA': 'mirna',
                'Target Gene': 'target_gene',
                'References (PMID)': 'ref_count'
            },
            axis=1
        )

        self.filename = "miRTarBase.{}.tsv".format(species.key)

        formatted_data.to_csv(
            self.filename,
            sep='\t',
            index=False,
            quoting=csv.QUOTE_NONE
        )

        return self.filename


class EnsemblTranscriptsFile(RefFile):
    def generate(self, biotype):
        self.filename = re.sub(r'\.fa$', '.{}.fa'.format(biotype), self.src_name)

        with open(self.src_name) as fa_in:
            fa_txt = fa_in.read()

        with open(self.filename, 'w') as out:
            for fa_id, fa_seq in parse_fasta(fa_txt):
                m = re.search(r'{}'.format(biotype), fa_id)
                if m:
                    print(">{}\n{}".format(fa_id, fa_seq), file=out)

        return self.filename


class RepChrM(RefFile):
    def generate(self):
        file_path = os.path.dirname(self.src_name)
        self.filename = os.path.join(file_path, "RepChrM.fa")

        with open(self.src_name) as fa_in:
            fa_txt = fa_in.read()

        with open(self.filename, 'w') as out:
            for fa_id, fa_seq in parse_fasta(fa_txt):
                m = re.search(r'^(?:chr)?MT?', fa_id)
                if m:
                    print(">{}\n{}\n{}".format(fa_id, fa_seq, fa_seq),
                          file=out)

        return self.filename


class OtherTranscripts:
    def __init__(self, src_type, pc_src, lncRNA_src, repChrM_src):
        self.src_type = src_type
        self.pc_src = pc_src
        self.lncRNA_src = lncRNA_src
        self.repChrM_src = repChrM_src
        self.filename = 'others.fa'

    def generate(self):
        if self.src_type == 'ensembl':
            pc_ref = EnsemblTranscriptsFile(self.pc_src.filename)
            pc_ref.generate('protein_coding')

            lncRNA_ref = EnsemblTranscriptsFile(self.lncRNA_src.filename)
            lncRNA_ref.generate('lncRNA')

        elif self.src_type == 'gencode':
            pc_ref = self.pc_src
            lncRNA_ref = self.lncRNA_src

        repChrM_ref = RepChrM(self.repChrM_src.filename)
        repChrM_ref.generate()

        with open(self.filename, 'wb') as out:
            for ref in [pc_ref, lncRNA_ref, repChrM_ref]:
                with open(ref.filename, 'rb') as f_in:
                    shutil.copyfileobj(f_in, out)

        return self.filename


class Files:
    def __init__(self, files):
        self.files = files

    def download(self):
        for file_ in self.files:
            file_.download()

    def unzip(self):
        for file_ in self.files:
            file_.unzip()

    def __getitem__(self, n):
        return self.files[n]


class EnsemblFiles(Files):
    source = "ensembl"


class GencodeFiles(Files):
    source = "gencode"


class MirTargetRef:
    def __init__(self, ref_files, ref_names, species):
        self.ref_files = ref_files
        self.ref_names = ref_names
        self.species = species
        self.filename = "mir_target.{}.tsv".format(self.species.key)

    def generate(self):
        merged_df = pd.DataFrame([], columns=['mirna', 'target_gene'])
        for ref_file, ref_name in zip(self.ref_files, self.ref_names):
            ref_df = pd.read_csv(ref_file.filename, sep='\t', dtype='object')
            ref_df = ref_df.pipe(self.add_ref_name, ref_name).pipe(self.add_ref_col, ref_name)
            merged_df = merged_df.merge(ref_df, on=['mirna', 'target_gene'], how="outer")

        for ref_name in self.ref_names:
            merged_df[ref_name].fillna('0', inplace=True)

        # promote the refname columns
        col_names = list(merged_df.columns)
        col_names = col_names[:2] + self.promote_items(col_names[2:], self.ref_names)
        merged_df = merged_df[col_names]

        merged_df.to_csv(self.filename, sep='\t', index=False)

    @staticmethod
    def add_ref_name(ref_df, ref_name, sep='__'):
        data_cols = ref_df.columns[2:]
        new_col_names = dict(map(lambda col: (col, "{}{}{}".format(ref_name, sep, col)), data_cols))
        return ref_df.rename(new_col_names, axis=1)

    @staticmethod
    def add_ref_col(ref_df, ref_name):
        return ref_df.assign(**{ref_name: '1'})

    @staticmethod
    def promote_items(all_items, to_be_promoted):
        result_items = all_items.copy()
        for item in to_be_promoted[::-1]:
            if item in result_items:
                item_idx = result_items.index(item)
                result_items = [result_items[item_idx]] + result_items[:item_idx] + result_items[(item_idx + 1):]
        return result_items


def generate(species, source, version, ref_dir):
    with cwd(ref_dir):
        species = species_list[species]

        if source == "ensembl":
            anno_file = rs.EnsemblAnnotation(species.name, version)
            genome_file = rs.EnsemblGenome(species.name, version)
            other_transcripts_files = EnsemblFiles(
                [
                    rs.EnsemblCDna(species.name, version),
                    rs.EnsemblNCRna(species.name, version)
                ]
            )

        elif source == "gencode":

            if species.key == 'hsa':
                species_key = 'human'
            elif species.key == 'mmu':
                species_key = 'mouse'

            anno_file = rs.GencodeAnnotation(species_key, version)
            genome_file = rs.GencodeGenome(species_key, version)
            other_transcripts_files = GencodeFiles(
                [
                    rs.GencodeProteinCodingTranscripts(species_key, version),
                    rs.GencodeLongNonCodingTranscripts(species_key, version)
                ]
            )

        elif source.startswith("ensembl_"):
            field = source.split('_')[1]
            anno_file = rs.EnsemblSisterAnnotation(field, species.name, version)
            genome_file = rs.EnsemblSisterGenome(field, species.name, version)
            other_transcripts_files = EnsemblFiles(
                [
                    rs.EnsemblSisterCDna(field, species.name, version),
                    rs.EnsemblSisterNCRna(field, species.name, version)
                ]
            )

        else:
            raise rs.SourceNotSupportError(source)

        mir_seq_file = rs.MiRBaseMiRNA(None, "22")
        mir_target_files = Files(
            [
                rs.MiRTarBaseResource(None, "7.0"),
                rs.MiRDBData(species.key, "6.0")
            ]
        )

        # download
        anno_file.download()
        genome_file.download()
        mir_seq_file.download()

        mir_target_files.download()
        other_transcripts_files.download()

        # unzip
        genome_file.unzip()
        mir_seq_file.unzip()

        mir_target_files.unzip()
        other_transcripts_files.unzip()

        # genref
        anno_ref = AnnoRef(anno_file.filename)
        anno_ref.generate()

        mir_ref = MirRef(mir_seq_file.filename)
        mir_ref.generate(species.key)

        # miRTarBase
        miRTarBase_ref = MiRTarBaseRef(mir_target_files[0].filename)
        miRTarBase_ref.generate(species)

        # miRNAs updater
        updater = MatureMiRNAUpdater("21", "22", species.key)
        updater.create()

        updated_file = "miRTarBase.{}.miRBase_v22.tsv".format(species.key)
        updater.update_file(
            miRTarBase_ref.filename,
            updated_file,
            col_key=0,
            inplace=True,
            remove_deleted=True
        )
        miRTarBase_ref.filename = updated_file

        mir_target_refs = [
            miRTarBase_ref,
            mir_target_files[1]
        ]
        mir_target_ref = MirTargetRef(
            mir_target_refs,
            [
                "miRTarBase",
                "miRDB"
            ],
            species
        )
        mir_target_ref.generate()

        others_ref = OtherTranscripts(
            other_transcripts_files.source,
            *other_transcripts_files.files,
            genome_file
        )
        others_ref.generate()

        # config
        info = {
            'species': species.key,
            'source': source,
            'version': anno_file.version
        }

        ref_files = {
            'anno_db': anno_ref.filename,
            'ref_file': genome_file.filename,
            'mir_ref': mir_ref.filename,
            'mir_target': mir_target_ref.filename,
            'other_transcripts': others_ref.filename
        }

        return info, ref_files



from collections import namedtuple

Species = namedtuple("Species", ['key', 'name', 'fullname', 'tax_id'])

species_list = {
    'ath': Species('ath', 'arabidopsis_thaliana', 'Arabidopsis thaliana', '3702'),
    'bmo': Species('bmo', 'bombyx_mori', 'Bombyx mori', '7091'),
    'bta': Species('bta', 'bos_taurus', 'Bos taurus', '9913'),
    'cel': Species('cel', 'caenorhabditis_elegans', 'Caenorhabditis elegans', '6239'),
    'cfa': Species('cfa', 'canis_familiaris', 'Canis familiaris', '9615'),
    'dre': Species('dre', 'danio_rerio', 'Danio rerio', '7955'),
    'dme': Species('dme', 'drosophila_melanogaster', 'Drosophila melanogaster', '7227'),
    'gga': Species('gga', 'gallus_gallus', 'Gallus gallus', '9031'),
    'hsa': Species('hsa', 'homo_sapiens', 'Homo sapiens', '9606'),
    'mmu': Species('mmu', 'mus_musculus', 'Mus musculus', '10090'),
    'osa': Species('osa', 'oryza_sativa', 'Oryza sativa', '4530'),
    'ola': Species('ola', 'oryzias_latipes', 'Oryzias latipes', '8090'),
    'oar': Species('oar', 'ovis_aries', 'Ovis aries', '9940'),
    'rno': Species('rno', 'rattus_norvegicus', 'Rattus norvegicus', '10116'),
    'ssc': Species('ssc', 'sus_scrofa', 'Sus scrofa', '9823'),
    'tgu': Species('tgu', 'taeniopygia_guttata', 'Taeniopygia guttata', '59729'),
    'xtr': Species('xtr', 'xenopus_tropicalis', 'Xenopus tropicalis', '8364')
}



import copy
import random
from lxml import etree


class CyXGMML:
    NSMAP = {
        "dc": "http://purl.org/dc/elements/1.1/",
        "xlink": "http://www.w3.org/1999/xlink",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "cy": "http://www.cytoscape.org",
        None: "http://www.cs.rpi.edu/XGMML"
    }
    TYPE = {
        str: 'string',
        bool: 'boolean',
        list: 'list',
        int: 'integer'
    }

    def __init__(self, graph):
        self.create_root()
        self.load(graph)

    def create_root(self):
        attrs = {
            "id": str(random.randint(0, 5000)),
            "label": "circmimi_result",
            "directed": "1",
            "{http://www.cytoscape.org}documentVersion": "3.0"
        }
        self.root = etree.Element(
            'graph',
            attrib=attrs,
            nsmap=self.NSMAP
        )

    def load(self, graph):

        self.update_attrs(self.root, **graph.graph)

        # add nodes
        for node, attrs in graph.nodes(True):
            attrs = copy.deepcopy(attrs)
            id_ = attrs.pop('id', node)
            label = attrs.pop('label', node)
            data = attrs.pop('data', {})
            data_type = attrs.pop('data_type', "")
            xgmml_node = self.add_node(self.root, id_, label, **attrs)

            self.add_att(xgmml_node, 'data_type', data_type)

            for k, v in data.items():
                self.add_att(xgmml_node, k, v)

        # add edges
        for source, target in graph.edges:
            attrs = copy.deepcopy(graph.edges[source, target])
            id_ = attrs.pop('id', "")
            label = attrs.pop('label', "{} - {}".format(source, target))
            data = attrs.pop('data', {})
            data_type = attrs.pop('data_type', "")
            xgmml_edge = self.add_edge(
                self.root,
                id_,
                label,
                source,
                target,
                **attrs
            )

            self.add_att(xgmml_edge, 'data_type', data_type)
            self.add_att(xgmml_edge, 'source', source)
            self.add_att(xgmml_edge, 'target', target)

            for k, v in data.items():
                self.add_att(xgmml_edge, k, v)

    def to_string(self):
        string = etree.tostring(
            self.root,
            xml_declaration=True,
            pretty_print=True,
            encoding='utf-8',
            standalone=True
        )

        return string

    @staticmethod
    def add_node(parent, id_, label, **attrs):
        node = etree.SubElement(
            parent,
            'node',
            id=id_,
            label=label,
            **attrs
        )

        return node

    @staticmethod
    def add_edge(parent, id_, label, source, target, **attrs):
        edge = etree.SubElement(
            parent,
            'edge',
            id=id_,
            label=label,
            source=source,
            target=target,
            **attrs
        )

        return edge

    @classmethod
    def add_att(cls, parent, name, value, **attrs):
        type_ = type(value)
        if type_ is int:
            value = str(value)
        elif type_ is bool:
            value = str(int(value))

        att = etree.SubElement(
            parent,
            'att',
            name=name,
            value=value,
            type=cls.TYPE[type_],
            **attrs
        )
        return att

    @staticmethod
    def add_graphics(parent, **attrs):
        graphics = etree.SubElement(parent, 'graphics', **attrs)
        return graphics

    @staticmethod
    def update_attrs(elt, **attrs):
        for k, v in attrs.items():
            elt.set(k, v)


def write_xgmml(graph, path):
    xgmml = CyXGMML(graph)

    with open(path, 'wb') as out:
        out.write(xgmml.to_string())



import gzip
import re
from operator import itemgetter
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from circmimi.models import (Base, Chromosome, Strand, Biotype, Gene,
                             Transcript, Exon, TranscriptExon, DonorSite,
                             AcceptorSite)


class RegionAttr:
    attr_patter = re.compile(r'([^ ;]+) \"?([^;"]+)\"?;')

    def __init__(self, attr_string):
        self._attrs = dict(re.findall(self.attr_patter, attr_string))
        self._fix_attrs_from_ensembl()

    def _fix_attrs_from_ensembl(self):
        if "gene_biotype" in self._attrs:
            self._attrs["gene_type"] = self._attrs["gene_biotype"]

        if "transcript_biotype" in self._attrs:
            self._attrs["transcript_type"] = self._attrs["transcript_biotype"]

        if "gene_version" in self._attrs:
            self._attrs["gene_id"] = "{}.{}".format(
                self._attrs["gene_id"],
                self._attrs["gene_version"]
            )

        if "transcript_version" in self._attrs:
            self._attrs["transcript_id"] = "{}.{}".format(
                self._attrs["transcript_id"],
                self._attrs["transcript_version"]
            )

    def get(self, *attr_names):
        return [self._attrs.get(attr) for attr in attr_names]


class GetSite:
    get_start = itemgetter(0, 1, 3)
    get_end = itemgetter(0, 2, 3)

    @classmethod
    def get_donor_site(cls, exon):
        if exon[3] == 1:
            return cls.get_end(exon)
        elif exon[3] == 2:
            return cls.get_start(exon)

    @classmethod
    def get_acceptor_site(cls, exon):
        if exon[3] == 1:
            return cls.get_start(exon)
        elif exon[3] == 2:
            return cls.get_end(exon)


class TablesRawData:
    def __init__(self):
        pass

    def parse(self, anno_file):
        if anno_file.endswith('.gz'):
            opened_file = gzip.open(anno_file, 'rt')
        elif anno_file.endswith('.gtf'):
            opened_file = open(anno_file)
        else:
            raise Exception('File format not supported!')

        # Parse annotation gtf
        genes = []
        transcripts = []
        exons_tmp = []
        with opened_file as gz_in:
            for line in gz_in:
                if not line.startswith('#'):
                    data = line.rstrip('\n').split('\t')
                    region_type = data[2]
                    if region_type in ['gene', 'transcript', 'exon']:
                        attrs = RegionAttr(data[8])
                        if region_type == 'gene':
                            genes.append(
                                attrs.get(
                                    'gene_id',
                                    'gene_name',
                                    'gene_type'
                                )
                            )
                        elif region_type == 'transcript':
                            transcripts.append(
                                attrs.get(
                                    'transcript_id',
                                    'gene_id',
                                    'transcript_type'
                                )
                            )
                        elif region_type == 'exon':
                            exons_tmp.append(
                                list(itemgetter(0, 3, 4, 6)(data)) +
                                attrs.get(
                                    'transcript_id',
                                    'exon_number'
                                )
                            )

        chromosomes = sorted(set(map(itemgetter(0), exons_tmp)))
        strands = sorted(set(map(itemgetter(3), exons_tmp)))
        biotypes = sorted(set(map(itemgetter(2), genes + transcripts)))

        chromosomes_dict = self._get_index_map(chromosomes)
        strands_dict = self._get_index_map(strands)
        biotypes_dict = self._get_index_map(biotypes)
        genes_dict = self._get_index_map(map(itemgetter(0), genes))
        transcripts_dict = self._get_index_map(map(itemgetter(0), transcripts))

        # replace values by index number
        for g in genes:
            g[2] = biotypes_dict[g[2]]

        for t in transcripts:
            t[1] = genes_dict[t[1]]
            t[2] = biotypes_dict[t[2]]

        for e in exons_tmp:
            e[0] = chromosomes_dict[e[0]]
            e[1] = int(e[1])
            e[2] = int(e[2])
            e[3] = strands_dict[e[3]]
            e[4] = transcripts_dict[e[4]]
            e[5] = int(e[5])

        # get exons data
        exons = sorted(set(map(itemgetter(0, 1, 2, 3), exons_tmp)))
        exons_dict = self._get_index_map(exons)

        # get exon_transcript relation data
        exon_transcript = [[e[4], e[5], exons_dict[tuple(e[:4])]]
                           for e in exons_tmp]

        # get donor & acceptor
        donor_sites = sorted(set(map(GetSite.get_donor_site, exons)))
        acceptor_sites = sorted(set(map(GetSite.get_acceptor_site, exons)))

        donor_sites_dict = self._get_index_map(donor_sites)
        acceptor_sites_dict = self._get_index_map(acceptor_sites)

        # append the donor and acceptor indices to exons data
        exons_with_da = [list(e) +
                         [donor_sites_dict[GetSite.get_donor_site(e)],
                          acceptor_sites_dict[GetSite.get_acceptor_site(e)]]
                         for e in exons]

        self.chromosomes = chromosomes
        self.strands = strands
        self.biotypes = biotypes
        self.genes = genes
        self.transcripts = transcripts
        self.exons = exons_with_da
        self.exon_transcript = exon_transcript
        self.donor_sites = donor_sites
        self.acceptor_sites = acceptor_sites

    @staticmethod
    def _get_index_map(keys):
        return {k: i for i, k in enumerate(keys, start=1)}


def _write_data_to_db(session, data, DataModal):
    for row in data:
        if type(row) == str:
            session.add(DataModal(row))
        else:
            session.add(DataModal(*row))
    session.commit()


def generate(gtf_path, db_path):
    engine = create_engine('sqlite:///{}'.format(db_path))

    # create tables
    Base.metadata.create_all(bind=engine)

    Session = sessionmaker(bind=engine)
    session = Session()

    # parse raw data
    tables_raw_data = TablesRawData()
    tables_raw_data.parse(gtf_path)

    # write raw data to db
    _write_data_to_db(session, tables_raw_data.chromosomes, Chromosome)
    _write_data_to_db(session, tables_raw_data.strands, Strand)
    _write_data_to_db(session, tables_raw_data.biotypes, Biotype)
    _write_data_to_db(session, tables_raw_data.genes, Gene)
    _write_data_to_db(session, tables_raw_data.transcripts, Transcript)
    _write_data_to_db(session, tables_raw_data.exons, Exon)
    _write_data_to_db(session, tables_raw_data.exon_transcript, TranscriptExon)
    _write_data_to_db(session, tables_raw_data.donor_sites, DonorSite)
    _write_data_to_db(session, tables_raw_data.acceptor_sites, AcceptorSite)



#! /usr/bin/env python
from __future__ import print_function

import argparse
import subprocess as sp
import os
import sys
import shutil
import random
import time
import re


def mp_blat(reference_file, fasta_file, output_file, num_of_process=1, tmp_path=".", blat_bin="blat", blat_opt=""):

    blat_opt = re.findall("([^ ]+)", blat_opt)

    if num_of_process == 1:
        # do blat directly
        sp.call([blat_bin, reference_file, fasta_file, output_file] + blat_opt)
    elif num_of_process > 1:

        # split file

        print("Start to split the input into {} part ... ".format(num_of_process), file=sys.stderr)
        split_start = time.time()

        fasta_file_size = os.stat(fasta_file).st_size
        with open(fasta_file) as fa_data:

            # get the number of all reads and their start pos in this file
            all_reads_pos = []
            while fa_data.tell() < fasta_file_size:
                next_char = fa_data.read(1)
                if next_char == ">":
                    all_reads_pos.append(fa_data.tell()-1)
            num_of_reads = len(all_reads_pos)

            fa_data.seek(0)

            if num_of_reads < num_of_process:
                splice_positions = all_reads_pos[1:]
            else:
                # eg. 70/20 = 3 ... 10
                #   ----> [3, 3, 3, ..., 3, 3, ..., 3]
                #   ----> [4, 4, 4, ..., 4, 3, ..., 3]
                #          ^^^^^^^^^^^^^^^
                #                10
                #          ^^^^^^^^^^^^^^^^^^^^^^^^^^
                #                      20
                #         
                #   ----> [4, 8, 12, ..., 40, 43, ..., 67]
                #          ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
                #                       19
                #
                #  Apply the index to "all_reads_pos"
                #   ----> [x[4], x[8], ..., x[67]]
                #   ----> [x[0]] + [x[4], x[8], ..., x[67]] + [file_size]
                #
                #  Get the size for each file
                #   ----> [x[4] - x[0], x[8] - x[4], ..., file_size - x[67]]
                #

                num_of_reads_for_each_blat = [num_of_reads // num_of_process]*num_of_process
                for i in range(num_of_reads % num_of_process):
                    num_of_reads_for_each_blat[i] += 1

                splice_positions = []
                idx = 0
                for n in num_of_reads_for_each_blat[:-1]:
                    idx += n
                    splice_positions.append(all_reads_pos[idx])

            edge_pos = [all_reads_pos[0]] + splice_positions + [fasta_file_size]
            read_size_for_each_blat = [edge_pos[1:][i] - edge_pos[:-1][i] for i in range(len(edge_pos)-1)]

            del all_reads_pos

            # create tmp files
            tmp_dir = "{}/mp_blat.tmp.{}".format(tmp_path.rstrip('/'), str(random.random())[2:8])
            os.mkdir(tmp_dir)

            tmp_fa_files = []
            for i, size in enumerate(read_size_for_each_blat):
                tmp_file_name = "{}/blat_tmp.part_{}.fa".format(tmp_dir, i+1)

                tmp_file = open(tmp_file_name, 'w')
                tmp_file.write(fa_data.read(size))
                tmp_file.close()

                tmp_fa_files.append(tmp_file_name)

        split_stop = time.time()
        print("Time cost of split_file = {} sec".format(split_stop - split_start), file=sys.stderr)


        # multi-processing blat

        print("Start to do the mp_blat using {} processes ... ".format(num_of_process), file=sys.stderr)
        blat_start = time.time()

        tmp_res_files = []
        for i, fa_file in enumerate(tmp_fa_files):
            tmp_file_name = "{}/blat_tmp.part_{}.psl".format(tmp_dir, i+1)
            tmp_res_files.append(tmp_file_name)

        proc = []
        for i in range(len(tmp_fa_files)):
            proc.append(sp.Popen([blat_bin, reference_file, tmp_fa_files[i], tmp_res_files[i]] + blat_opt))

        for p in proc:
            p.wait()

        blat_stop = time.time()
        print("Time cost of mp_blat = {} sec".format(blat_stop - blat_start), file=sys.stderr)

        # merge results
        print("Merging results ... ", file=sys.stderr)
        merge_start = time.time()

        with open(output_file, 'w') as output:
            all_res_files = [open(res_file) for res_file in tmp_res_files]

            # titles
            for i in range(5):
                output.write(all_res_files[0].readline())

            after_title_pos = all_res_files[0].tell()

            for res_file in all_res_files:
                res_file.seek(after_title_pos)
                output.write(res_file.read())
                res_file.close()

        merge_stop = time.time()
        print("Time cost of merge_result = {} sec".format(merge_stop - merge_start), file=sys.stderr)

        # clean tmp files and dir
        shutil.rmtree(tmp_dir)


def cli():
    parser = argparse.ArgumentParser()
    parser.add_argument("reference")
    parser.add_argument("fasta")
    parser.add_argument("output")
    parser.add_argument("-p", "--process", type=int, default=1)
    parser.add_argument("--tmp_path", type=str, default=".")
    parser.add_argument("--blat_bin", default="blat")
    parser.add_argument("--blat_opt", type=str, default="")
    args = parser.parse_args()

    start = time.time()
    mp_blat(args.reference, args.fasta, args.output, args.process, args.tmp_path, args.blat_bin, args.blat_opt)
    stop = time.time()
    print("Total time cost = {} sec".format(stop - start), file=sys.stderr)


if __name__ == "__main__":
    cli()



import os
import configparser


DEFAULT_REF_CONFIG = 'refs.cfg'

CONFIG_TEMPL = """
    [info]
    species =
    source =
    version =

    [refs]
    anno_db =
    ref_file =
    mir_ref =
    mir_target =
    other_transcripts =
"""


class RefConfig:
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.config.read_string(CONFIG_TEMPL)

    def read(self, cfg_file):
        self.config.read(cfg_file)

    def write(self, ref_dir, cfg_file_name=DEFAULT_REF_CONFIG):
        cfg_file = os.path.join(ref_dir, cfg_file_name)
        with open(cfg_file, 'w') as config_file:
            self.config.write(config_file)

    def __getitem__(self, key):
        return self.config[key]


def get_refs(ref_dir):
    cfg_file = os.path.join(ref_dir, DEFAULT_REF_CONFIG)

    configObj = RefConfig()
    configObj.read(cfg_file)
    config = configObj.config

    anno_db = os.path.join(ref_dir, config['refs']['anno_db'])
    ref_file = os.path.join(ref_dir, config['refs']['ref_file'])
    mir_ref = os.path.join(ref_dir, config['refs']['mir_ref'])
    mir_target = os.path.join(ref_dir, config['refs']['mir_target'])
    other_transcripts = os.path.join(ref_dir, config['refs']['other_transcripts'])

    return anno_db, ref_file, mir_ref, mir_target, other_transcripts



import tempfile as tp
import subprocess as sp
import re
import pandas as pd


class Seq:
    @classmethod
    def get_seq(cls, bed_df, ref_file):
        with tp.NamedTemporaryFile(dir='.') as tmp_bed_file:

            with open(tmp_bed_file.name, 'w') as bed_out:
                bed_str = bed_df.to_csv(sep='\t', header=False, index=False)
                bed_out.write(bed_str)

            fasta_txt = get_fasta(tmp_bed_file.name, ref_file, use_blocks=True)

            fasta_df = pd.DataFrame(
                parse_fasta(fasta_txt),
                columns=['name', 'seq']
            )

            return fasta_df

    @staticmethod
    def to_fasta(fasta_df):
        fa_txt = ""
        for name, seq in fasta_df.values:
            fa_txt += ">{}\n{}\n".format(name, seq)
        return fa_txt

    @staticmethod
    def extend_seq_for_circ_js(s):
        seq = s['seq']
        extended_seq = seq + seq[:30]
        return extended_seq

    @classmethod
    def get_extended_seq(cls, bed_df, ref_file):
        fasta_df = cls.get_seq(bed_df, ref_file)
        if not fasta_df.empty:
            fasta_df['seq'] = fasta_df.apply(
                cls.extend_seq_for_circ_js,
                axis=1
            )

        return fasta_df


def get_fasta(bed_file, ref_file, use_blocks=False, bedtools_bin='bedtools'):
    cmd = [bedtools_bin, 'getfasta']
    cmd += ['-fi', ref_file]
    cmd += ['-bed', bed_file]
    cmd += ['-fo', '/dev/stdout']
    cmd += ['-name', '-s']

    if use_blocks:
        cmd += ['-split']

    res = sp.run(cmd, stdout=sp.PIPE, encoding='utf-8')

    return res.stdout


def parse_fasta(fasta_txt):
    for m in re.finditer(r">(.+?)(?:\([+-]\))?\n([^>]+)", fasta_txt):
        fa_id = m.group(1)
        fa_seq = ''.join(m.group(2).rstrip('\n').split('\n'))
        yield [fa_id, fa_seq]



import os
import shutil
import subprocess as sp
import tempfile as tp
import random
import re
import pandas as pd
from itertools import cycle
from functools import reduce
from contextlib import contextmanager
from circmimi.seq import Seq


class Miranda:
    RESULT_TITLE = (
        'query_id',
        'reference_id',
        'score',
        'energy',
        'query_start',
        'query_end',
        'ref_start',
        'ref_end',
        'aln_length',
        'identity',
        'similarity',
        'aln_mirna',
        'aln_map',
        'aln_utr'
    )

    def __init__(self, ref_file, work_dir='.',
                 bin_path='miranda', options=None):
        self.ref_file = ref_file
        self.work_dir = work_dir
        self.bin_path = bin_path

        if options:
            self.options = options
        else:
            self.options = []
        self.options += ['-keyval']
        self.options = list(map(str, self.options))

    def _generate_cmd(self, seq_file, out_file=None):
        cmd = [self.bin_path, self.ref_file, seq_file] + self.options

        if out_file:
            cmd += ['-out', out_file]

        return cmd

    @staticmethod
    def _rand_id(n=5):
        return '{{:0>{}}}'.format(n).format(random.randrange(10**n))

    @contextmanager
    def _split_file(self, seq_file, num):
        tmp_dir = os.path.join(
            self.work_dir,
            'miranda_tmp_{}'.format(self._rand_id())
        )
        os.makedirs(tmp_dir)

        filenames = ['seq_file.part{}.fa'.format(i) for i in range(1, num + 1)]
        file_paths = [os.path.join(tmp_dir, fname) for fname in filenames]
        tmp_files = [open(path, 'w') for path in file_paths]

        tmp_files_iter = cycle(tmp_files)
        with open(seq_file) as seq_in:
            for line in seq_in:
                if line.startswith('>'):
                    tmp_f = next(tmp_files_iter)
                tmp_f.write(line)

        for tmp_f in tmp_files:
            tmp_f.close()

        try:
            yield file_paths
        finally:
            shutil.rmtree(tmp_dir)

    def run(self, seq_file, num_proc=1):
        if num_proc == 1:
            cmd = self._generate_cmd(seq_file)
            res = sp.run(cmd, stdout=sp.PIPE, encoding='utf-8')
            return res.stdout
        elif num_proc > 1:
            with self._split_file(seq_file, num=num_proc) as tmp_files:
                out_files = ["{}.result".format(tmp_f) for tmp_f in tmp_files]

                cmds = [self._generate_cmd(tmp_f, out_f)
                        for tmp_f, out_f in zip(tmp_files, out_files)]
                procs = [sp.Popen(cmd) for cmd in cmds]

                for p in procs:
                    p.wait()

                result = ''
                for out_f in out_files:
                    with open(out_f) as res_in:
                        result += res_in.read()

            return result

    @staticmethod
    def _get_value(res_line):
        return [kv.split('=')[1] for kv in res_line.split('\t')]

    @classmethod
    def parse_result(cls, raw_result):
        aln_res_pat = re.compile(r'^//hit_info\t(.*)', flags=re.M)
        for m in re.finditer(aln_res_pat, raw_result):
            yield cls._get_value(m.group(1))


def get_binding_sites(seq_df,
                      mir_ref_file,
                      work_dir='.',
                      num_proc=1,
                      miranda_options=None):

    if miranda_options is None:
        miranda_options = []

    miranda = Miranda(
        mir_ref_file,
        work_dir=work_dir,
        options=['-quiet'] + miranda_options
    )

    with tp.NamedTemporaryFile(dir=work_dir) as tmp_fa_file:
        with open(tmp_fa_file.name, 'w') as fa_out:
            fa_txt = Seq.to_fasta(seq_df)
            fa_out.write(fa_txt)

        raw_result = miranda.run(tmp_fa_file.name, num_proc=num_proc)
        miranda_df = pd.DataFrame(
            Miranda.parse_result(raw_result),
            columns=Miranda.RESULT_TITLE
        ).astype({
            'score': 'float',
            'energy': 'float',
            'query_start': 'int',
            'query_end': 'int',
            'ref_start': 'int',
            'ref_end': 'int',
            'aln_length': 'int',
            'identity': 'float',
            'similarity': 'float'
        })

        return miranda_df


class MirandaUtils:
    @staticmethod
    def append_exons_len(miranda_df, exons_len_df):
        miranda_df_with_total_len = miranda_df.merge(
            exons_len_df,
            left_on='reference_id',
            right_on='exons_id',
            how='left'
        ).drop('exons_id', axis=1)

        return miranda_df_with_total_len

    @classmethod
    def remove_redundant_result(cls, miranda_df):
        filtered_df = miranda_df.query('ref_start <= total_len')

        dropped_index = filtered_df.query(
            'ref_start == 1'
        ).assign(
            ext_ref_end=lambda df: df['ref_end'] + df['total_len']
        ).loc[:, [
            'query_id',
            'reference_id',
            'ext_ref_end'
        ]].reset_index(
        ).merge(
            filtered_df,
            left_on=['query_id', 'reference_id', 'ext_ref_end'],
            right_on=['query_id', 'reference_id', 'ref_end'],
            how='inner'
        ).loc[:, 'index']

        filtered_df = filtered_df.drop(dropped_index).reset_index(drop=True)

        return filtered_df

    @staticmethod
    def _is_cross_boundary(s):
        exons_len = s['total_len']
        ref_start = s['ref_start']
        ref_end = s['ref_end']

        if (ref_start <= exons_len) and (exons_len < ref_end):
            return '1'
        else:
            return '0'

    @classmethod
    def append_cross_boundary(cls, miranda_df_with_len):
        if miranda_df_with_len.empty:
            cross_boundary_res = pd.Series()
        else:
            cross_boundary_res = miranda_df_with_len.apply(
                cls._is_cross_boundary,
                axis=1
            )

        appended_res_df = miranda_df_with_len.assign(
            cross_boundary=cross_boundary_res
        )

        return appended_res_df

    @staticmethod
    def append_ev_id(df, exons_ev_id_df):
        df_with_ev_id = df.merge(
            exons_ev_id_df,
            left_on='reference_id',
            right_on='exons_id',
            how='left'
        ).drop(
            columns='exons_id'
        ).sort_values(
            ['ev_id', 'query_id', 'score'],
            ascending=[True, True, False]
        )

        return df_with_ev_id

    @staticmethod
    def append_merged_aln(miranda_df):
        miranda_df_with_merged_aln = miranda_df.assign(
            aln=lambda df: df['aln_mirna'] + df['aln_map'] + df['aln_utr']
        )

        return miranda_df_with_merged_aln

    @classmethod
    def generate_aln_id(cls, miranda_df_with_aln):
        all_uniq_alns = pd.Series(
            miranda_df_with_aln['aln'].unique(),
            name='aln'
        )
        miranda_df_with_aln_id = miranda_df_with_aln.merge(
            all_uniq_alns.reset_index(),
            on='aln',
            how='left'
        ).rename({'index': 'aln_id'}, axis=1)

        return miranda_df_with_aln_id

    @staticmethod
    def get_grouped_results(miranda_df_with_aln_id):
        grouped_res_df = miranda_df_with_aln_id[[
            'ev_id',
            'query_id',
            'score',
            'aln_id',
            'cross_boundary'
        ]].groupby([
            'ev_id',
            'query_id'
        ]).agg({
            'score': 'max',
            'aln_id': 'nunique',
            'cross_boundary': 'max'
        }).reset_index().rename(
            {
                'query_id': 'mirna',
                'score': 'max_score',
                'aln_id': 'count'
            },
            axis=1
        ).astype(
            {
                'ev_id': 'object',
                'mirna': 'object',
                'count': "Int64"
            }
        )

        return grouped_res_df



import os
import gzip
from contextlib import contextmanager


@contextmanager
def cwd(path):
    origin_pwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(origin_pwd)


def open_file(filename):
    if filename.endswith('.gz'):
        opened_file = gzip.open(filename, 'rt')
    else:
        opened_file = open(filename)

    return opened_file



import pandas as pd
import numpy as np
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from circmimi.models import Chromosome, Strand, DonorSite, AcceptorSite


class Annotation:
    def __init__(self, anno_db_file):
        self._anno_db_file = anno_db_file

        self.engine = create_engine('sqlite:///{}'.format(anno_db_file))
        self.Session = sessionmaker(bind=self.engine)
        self.session = self.Session()

        self.chr_dict = dict(
            self.session.query(Chromosome.name, Chromosome.id).all()
        )

        self.strand_dict = dict(
            self.session.query(Strand.name, Strand.id).all()
        )

    def _get_junc_site(self, JuncSiteType, chr_name, pos, strand=None):
        result = self.session\
            .query(JuncSiteType)\
            .filter_by(chr_id=self.chr_dict[chr_name], junc_site=pos)

        if strand is not None:
            result = result.filter_by(strand_id=self.strand_dict[strand])

        return result.first()

    def _get_nearest_junc_site(self, JuncSiteType, chr_name, pos, strand=None, dist=None):
        query_result = self.session\
            .query(JuncSiteType)\
            .filter_by(chr_id=self.chr_dict[chr_name])

        if strand is not None:
            query_result = query_result\
                .filter_by(strand_id=self.strand_dict[strand])

        up_result = query_result.filter(JuncSiteType.junc_site >= pos).first()
        down_result = query_result\
            .filter(JuncSiteType.junc_site <= pos)\
            .order_by(JuncSiteType.id.desc()).first()

        result = min([up_result, down_result],
                     key=lambda JuncSite: abs(JuncSite.junc_site - pos))

        if dist is not None:
            if abs(result.junc_site - pos) > dist:
                result = None

        return result

    def get_donor_site(self, chr_name, pos, strand=None):
        return self._get_junc_site(DonorSite, chr_name, pos, strand)

    def get_acceptor_site(self, chr_name, pos, strand=None):
        return self._get_junc_site(AcceptorSite, chr_name, pos, strand)

    def get_nearest_donor_site(self, chr_name, pos, strand=None, dist=None):
        return self._get_nearest_junc_site(DonorSite, chr_name, pos, strand, dist)

    def get_nearest_acceptor_site(self, chr_name, pos, strand=None, dist=None):
        return self._get_nearest_junc_site(AcceptorSite, chr_name, pos, strand, dist)

    @staticmethod
    def get_transcripts_of_exons(exons):
        transcripts = []
        for exon in exons:
            transcripts += exon.transcript

        return transcripts

    @classmethod
    def get_transcripts_of_js(cls, junc_site):
        if junc_site is None:
            return []

        exons = junc_site.exons
        transcripts = cls.get_transcripts_of_exons(exons)
        return transcripts

    @staticmethod
    def get_inter_exons_data(transcript, donor, acceptor):
        transcript.init_exons()

        exon_donor = transcript.get_exon_by_donor(donor)
        exon_acceptor = transcript.get_exon_by_acceptor(acceptor)

        exon_number_donor = transcript.get_exon_number(exon_donor)
        exon_number_acceptor = transcript.get_exon_number(exon_acceptor)

        inter_exons = tuple(
            transcript.get_inter_exons(exon_number_donor, exon_number_acceptor)
        )

        anno_data = [transcript,
                     exon_number_donor,
                     exon_number_acceptor,
                     inter_exons]

        return anno_data


class Annotator:
    _CHECK_LIST = [
        'donor site not at the annotated boundary',
        'acceptor site not at the annotated boundary',
        'donor/acceptor sites not at the same transcript isoform'
    ]

    def __init__(self, anno_db_file):
        self._db = Annotation(anno_db_file)

    def _report_status(self, ev_id, status=None):
        if ev_id in self._checking_result.index:
            if status is not None:
                self._checking_result.loc[ev_id, status] = '1'
        else:
            ev_status = pd.Series(
                ['0'] * len(self._CHECK_LIST),
                index=self._CHECK_LIST,
                name=ev_id
            )

            if status is not None:
                ev_status[status] = '1'

            self._checking_result = self._checking_result.append(ev_status)

    def _get_anno_data_of_ev(self, s):
        chr_, donor_site, acceptor_site, strand = \
            s[['chr', 'donor', 'acceptor', 'strand']]

        ev_id = s.name

        donor = self._db.get_donor_site(chr_, donor_site, strand)
        acceptor = self._db.get_acceptor_site(chr_, acceptor_site, strand)

        donor_transcripts = self._db.get_transcripts_of_js(donor)
        acceptor_transcripts = self._db.get_transcripts_of_js(acceptor)

        common_transcripts = sorted(
            set(donor_transcripts) & set(acceptor_transcripts),
            key=lambda transcript: transcript.transcript_id
        )

        transcripts_data = [
            [ev_id] + self._db.get_inter_exons_data(transcript, donor, acceptor)
            for transcript in common_transcripts
        ]

        df_cols = [
            'ev_id',
            'transcript',
            'exon_number_d',
            'exon_number_a',
            'exons'
        ]

        transcripts_data_df = pd.DataFrame(transcripts_data, columns=df_cols)

        if common_transcripts == []:
            if donor is None:
                self._report_status(ev_id, self._CHECK_LIST[0])

            if acceptor is None:
                self._report_status(ev_id, self._CHECK_LIST[1])

            if (donor is not None) and (acceptor is not None):
                self._report_status(ev_id, self._CHECK_LIST[2])
        else:
            self._report_status(ev_id)

        return transcripts_data_df

    def annotate(self, df):
        self._checking_result = pd.DataFrame(
            [],
            columns=self._CHECK_LIST
        ).rename_axis('ev_id')

        if df.empty:
            anno_df = pd.DataFrame(
                [],
                columns=[
                    'ev_id',
                    'transcript',
                    'exon_number_d',
                    'exon_number_a',
                    'exons'
                ]
            )
        else:
            raw_anno_dfs = df.apply(self._get_anno_data_of_ev, axis=1)
            anno_df = pd.concat(list(raw_anno_dfs)).reset_index(drop=True)

        return anno_df, self._checking_result



import os.path
import pandas as pd
from circmimi.circ import CircEvents
from circmimi.bed import BedUtils
from circmimi.seq import Seq
from circmimi.miranda import get_binding_sites, MirandaUtils


class Circmimi:
    def __init__(self,
                 anno_db_file,
                 ref_file,
                 mir_ref_file,
                 mir_target_file,
                 other_ref_file=None,
                 work_dir='.',
                 num_proc=1,
                 miranda_options=None):

        self.anno_db_file = anno_db_file
        self.ref_file = ref_file
        self.mir_ref_file = mir_ref_file
        self.mir_target_file = mir_target_file
        self.other_ref_file = other_ref_file
        self.work_dir = work_dir
        self.num_proc = num_proc
        self.miranda_options = miranda_options

        self.circ_events = None
        self.uniq_exons_df = None
        self.bed_df = None
        self.seq_df = None
        self.miranda_df = None
        self.grouped_res_df = None

    def run(self, circ_file):

        self.circ_events = CircEvents(circ_file)
        self.circ_events.check_annotation(self.anno_db_file)

        if self.other_ref_file is not None:
            self.circ_events.check_ambiguous(
                self.anno_db_file,
                self.ref_file,
                self.other_ref_file,
                work_dir=self.work_dir,
                num_proc=self.num_proc
            )

        self.uniq_exons_df = self.circ_events.clear_anno_df.pipe(self._get_uniq_exons)
        self.bed_df = self.uniq_exons_df.pipe(
            BedUtils.to_regions_df
        ).pipe(
            BedUtils.to_bed_df
        )

        self.seq_df = self.bed_df.pipe(Seq.get_extended_seq, ref_file=self.ref_file)

        self.miranda_df = self.seq_df.pipe(
            get_binding_sites,
            mir_ref_file=self.mir_ref_file,
            work_dir=self.work_dir,
            num_proc=self.num_proc,
            miranda_options=self.miranda_options
        ).pipe(
            MirandaUtils.append_exons_len,
            exons_len_df=self.uniq_exons_df[['exons_id', 'total_len']]
        ).pipe(
            MirandaUtils.remove_redundant_result
        ).pipe(
            MirandaUtils.append_cross_boundary
        ).pipe(
            MirandaUtils.append_ev_id,
            exons_ev_id_df=self.uniq_exons_df[['exons_id', 'ev_id']]
        ).pipe(
            MirandaUtils.append_merged_aln
        ).pipe(
            MirandaUtils.generate_aln_id
        ).drop('aln', axis=1)

        self.grouped_res_df = MirandaUtils.get_grouped_results(self.miranda_df)

        self.mir_target_db = get_mir_target_db(self.mir_target_file)

        self.gene_symbol_df = self.circ_events.clear_anno_df.assign(
            host_gene=lambda df: df['transcript'].apply(lambda t: t.gene.gene_symbol)
        ).loc[:, ['ev_id', 'host_gene']].drop_duplicates().reset_index(drop=True)

    def get_result_table(self):
        res_df = self.circ_events.clear_df.reset_index().merge(
            self.gene_symbol_df,
            on='ev_id',
            how='inner'
        ).merge(
            self.grouped_res_df,
            on='ev_id',
            how='inner'
        ).merge(
            self.mir_target_db,
            on='mirna',
            how='inner'
        )

        circ_miRNA_count = res_df[['ev_id', 'mirna']].drop_duplicates().rename(
            {
                'mirna': '#circRNA_miRNA'
            },
            axis=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int')
        self.circ_events.submit_to_summary(circ_miRNA_count, type_='summary')

        circ_mRNA_count = res_df[['ev_id', 'target_gene']].drop_duplicates().rename(
            {
                'target_gene': '#circRNA_mRNA'
            },
            axis=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int')
        self.circ_events.submit_to_summary(circ_mRNA_count, type_='summary')

        total_count = res_df[['ev_id']].assign(
            circRNA_miRNA_mRNA=1
        ).groupby(
            'ev_id'
        ).agg(
            'count'
        ).pipe(
            self.circ_events.expand_to_all_events,
            fillna_value=0
        ).astype('int').rename(
            {
                'circRNA_miRNA_mRNA': '#circRNA_miRNA_mRNA'
            },
            axis=1
        )
        self.circ_events.submit_to_summary(total_count, type_='summary')

        return res_df.drop('ev_id', axis=1)

    def save_circRNAs_summary(self, out_file):
        self.circ_events.get_summary().to_csv(out_file, sep='\t', index=False)

    def save_clear_circRNAs(self, out_file):
        self.circ_events.clear_df.to_csv(
            out_file,
            sep='\t',
            index=False,
            header=False
        )

    @staticmethod
    def _get_total_length(list_of_obj):
        return sum(map(len, list_of_obj))

    @classmethod
    def _get_uniq_exons(cls, anno_df):
        if anno_df.empty:
            uniq_exons_df = pd.DataFrame(
                [],
                columns=['exons', 'ev_id', 'total_len', 'exons_id']
            )
        else:
            uniq_exons_df = anno_df[['exons', 'ev_id']]\
                .drop_duplicates()\
                .reset_index(drop=True)

            uniq_exons_df['total_len'] = uniq_exons_df.apply(
                lambda s: cls._get_total_length(s['exons']),
                axis=1
            )

            uniq_exons_df['exons_id'] = uniq_exons_df.apply(
                lambda s: "exons_{}".format(s.name),
                axis=1
            )

        return uniq_exons_df


def get_mir_target_db(mir_tar_db_path):
    db = pd.read_csv(mir_tar_db_path, sep='\t', dtype='object')

    assert list(db.columns[:2]) == ['mirna', 'target_gene'], \
        ("The column names of the first two columns"
         " should be 'mirna' and 'target_gene'")

    return db



import codecs
import os
import re
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


long_description = read('README.md')


setup(
    name='circmimi',
    version=find_version("circmimi", "__init__.py"),
    url='https://github.com/TreesLab/CircMiMi',
    packages=find_packages(),
    install_requires=[
        'click>=7.0',
        'sqlalchemy>=1.3.8',
        'numpy>=1.17.2',
        'pandas>=0.25.1',
        'xlrd>=1.2.0',
        'networkx>=2.4',
        'lxml>=4.5.0'
    ],
    entry_points={
        'console_scripts': [
            'circmimi_tools = circmimi.scripts.circmimi_tools:cli',
            'mp_blat.py = circmimi.scripts.mp_blat:cli'
        ]
    },
    description=("A toolset for investigating the interactions between"
                 " circRNA - miRNA - mRNA."),
    long_description=long_description,
    long_description_content_type="text/markdown"
)



from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Index, Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship


Base = declarative_base()


class Chromosome(Base):
    __tablename__ = 'chromosome'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Chromosome(name={})".format(self.name)


class Strand(Base):
    __tablename__ = 'strand'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return "Strand(name={})".format(self.name)


class Biotype(Base):
    __tablename__ = 'biotype'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return 'Biotype(type="{}")'.format(self.name)


class DonorSite(Base):
    __tablename__ = 'donor_site'

    id = Column(Integer, primary_key=True)
    chr_id = Column(Integer, ForeignKey('chromosome.id'))
    junc_site = Column(Integer)
    strand_id = Column(Integer, ForeignKey('strand.id'))
    chr_ = relationship("Chromosome")
    strand = relationship("Strand")
    exons = relationship("Exon", back_populates='donor')

    __table_args__ = (Index('donor_site_index', "chr_id", "junc_site"), )

    def __init__(self, chr_id, junc_site, strand_id):
        self.chr_id = chr_id
        self.junc_site = junc_site
        self.strand_id = strand_id

    def __repr__(self):
        return "DonorSite({}, {}, {})".format(
            self.chr_.name, self.junc_site, self.strand.name
        )

    @property
    def region(self):
        return [self.chr_.name, self.junc_site, self.strand.name]


class AcceptorSite(Base):
    __tablename__ = 'acceptor_site'

    id = Column(Integer, primary_key=True)
    chr_id = Column(Integer, ForeignKey('chromosome.id'))
    junc_site = Column(Integer)
    strand_id = Column(Integer, ForeignKey('strand.id'))
    chr_ = relationship("Chromosome")
    strand = relationship("Strand")
    exons = relationship("Exon", back_populates='acceptor')

    __table_args__ = (Index('acceptor_site_index', "chr_id", "junc_site"), )

    def __init__(self, chr_id, junc_site, strand_id):
        self.chr_id = chr_id
        self.junc_site = junc_site
        self.strand_id = strand_id

    def __repr__(self):
        return "AcceptorSite({}, {}, {})".format(
            self.chr_.name, self.junc_site, self.strand.name
        )

    @property
    def region(self):
        return [self.chr_.name, self.junc_site, self.strand.name]


class Exon(Base):
    __tablename__ = 'exon'

    id = Column(Integer, primary_key=True)
    chr_id = Column(Integer, ForeignKey('chromosome.id'))
    start = Column(Integer)
    end = Column(Integer)
    strand_id = Column(Integer, ForeignKey('strand.id'))
    donor_id = Column(Integer, ForeignKey('donor_site.id'), index=True)
    acceptor_id = Column(Integer, ForeignKey('acceptor_site.id'), index=True)

    chr_ = relationship("Chromosome")
    strand = relationship("Strand")
    donor = relationship("DonorSite", back_populates='exons')
    acceptor = relationship("AcceptorSite", back_populates='exons')
    _transcript_exon = relationship('TranscriptExon', back_populates='exon')

    def __init__(self, chr_id, start, end, strand_id, donor_id, acceptor_id):
        self.chr_id = chr_id
        self.start = start
        self.end = end
        self.strand_id = strand_id
        self.donor_id = donor_id
        self.acceptor_id = acceptor_id

    def __repr__(self):
        return "Exon({}, {}, {}, {})".format(
            self.chr_.name, self.start, self.end, self.strand.name
        )

    def __len__(self):
        return self.end - self.start + 1

    @property
    def transcript(self):
        return list(map(lambda t_e: t_e.transcript, self._transcript_exon))

    @property
    def region(self):
        return [self.chr_.name, self.start, self.end, self.strand.name]


class Gene(Base):
    __tablename__ = 'gene'

    id = Column(Integer, primary_key=True)
    gene_id = Column(String)
    gene_symbol = Column(String)
    gene_type_id = Column(Integer, ForeignKey('biotype.id'))
    gene_type = relationship('Biotype')
    transcripts = relationship('Transcript', back_populates='gene')

    def __init__(self, gene_id, gene_symbol, gene_type_id):
        self.gene_id = gene_id
        self.gene_symbol = gene_symbol
        self.gene_type_id = gene_type_id

    def __repr__(self):
        return 'Gene(gene_id="{}", gene_symbol="{}", gene_type="{}")'.format(
            self.gene_id, self.gene_symbol, self.gene_type.name
        )


class Transcript(Base):
    __tablename__ = 'transcript'

    id = Column(Integer, primary_key=True)
    transcript_id = Column(String)
    gid = Column(Integer, ForeignKey('gene.id'))
    transcript_type_id = Column(Integer, ForeignKey('biotype.id'))
    transcript_type = relationship('Biotype')
    gene = relationship('Gene', back_populates='transcripts')
    _transcript_exon = relationship('TranscriptExon',
                                    back_populates='transcript')

    def __init__(self, transcript_id, gid, transcript_type_id):
        self.transcript_id = transcript_id
        self.gid = gid
        self.transcript_type_id = transcript_type_id

    def __repr__(self):
        return 'Transcript(transcript_id="{}", biotype="{}")'.format(
            self.transcript_id, self.transcript_type.name
        )

    def init_exons(self):
        self.exons = list(map(lambda t_e: t_e.exon,
                              sorted(self._transcript_exon,
                                     key=lambda t_e: t_e.exon_number)))

    @staticmethod
    def _get_len_of_exons(exons):
        return sum(map(len, exons))

    def get_inter_exons(self, from_, to_):
        from_, to_ = sorted([from_, to_])
        return self.exons[(from_ - 1):((to_ - 1) + 1)]

    def get_len_of_inter_exons(self, from_, to_):
        return self._get_len_of_exons(self.get_inter_exons(from_, to_))

    @property
    def total_len_of_exons(self):
        return self._get_len_of_exons(self.exons)

    @property
    def total_number_of_exons(self):
        return len(self.exons)

    def get_nth_exon(self, n):
        if (n >= 1) and (n <= self.total_number_of_exons):
            return self.exons[n - 1]
        else:
            return None

    def get_exon_number(self, exon):
        return self.exons.index(exon) + 1

    def get_len_of_intron(self, exon_num_1, up_or_down, empty_value='NA'):
        if up_or_down == 'up':
            exon_num_2 = exon_num_1 - 1
        elif up_or_down == 'down':
            exon_num_2 = exon_num_1 + 1

        exon_1 = self.get_nth_exon(exon_num_1)
        exon_2 = self.get_nth_exon(exon_num_2)
        if exon_1 and exon_2:
            junc_sites = sorted([exon_1.start, exon_1.end,
                                 exon_2.start, exon_2.end])
            intron_start = junc_sites[1] + 1
            intron_end = junc_sites[2] - 1
            return intron_end - intron_start + 1
        else:
            return empty_value

    @property
    def is_protein_coding(self):
        return self.transcript_type.name == 'protein_coding'

    def _get_exon_by_junc_site(self, donor_or_acceptor):
        if isinstance(donor_or_acceptor, DonorSite):
            junc_site_type = 'donor'
        elif isinstance(donor_or_acceptor, AcceptorSite):
            junc_site_type = 'acceptor'

        filtered_exons = list(filter(
            lambda exon: getattr(exon, junc_site_type) == donor_or_acceptor,
            self.exons
        ))

        if len(filtered_exons) == 1:
            return filtered_exons[0]
        elif len(filtered_exons) == 0:
            raise Exception("No exon founded!")
        else:
            raise Exception(
                "More than one exon was founded! Please check your database!"
            )

    def get_exon_by_donor(self, donor):
        return self._get_exon_by_junc_site(donor)

    def get_exon_by_acceptor(self, acceptor):
        return self._get_exon_by_junc_site(acceptor)


class TranscriptExon(Base):
    __tablename__ = 'transcript_exon'

    tid = Column(Integer, ForeignKey('transcript.id'),
                 primary_key=True, index=True)
    exon_number = Column(Integer, primary_key=True)
    eid = Column(Integer, ForeignKey('exon.id'), index=True)

    transcript = relationship('Transcript', back_populates='_transcript_exon')
    exon = relationship('Exon', back_populates='_transcript_exon')

    def __init__(self, tid, exon_number, eid):
        self.tid = tid
        self.exon_number = exon_number
        self.eid = eid

    def __repr__(self):
        return "{} {} {}".format(
            self.transcript.transcript_id, self.exon_number, self.exon
        )



import pandas as pd
from circmimi.annotation import Annotator
from circmimi.ambiguous import AmbiguousChecker


class CircEvents:
    def __init__(self, filename):
        self._filename = filename
        self.original_df = self._read_file(self._filename)
        self.df = self._get_donor_acceptor_df(self.original_df)

        self._summary_columns = {
            'summary': [],
            'filters': []
        }

    @staticmethod
    def _read_file(filename):
        df = pd.read_csv(
            filename,
            sep='\t',
            names=['chr', 'pos1', 'pos2', 'strand'],
            dtype={
                'chr': 'category',
                'pos1': 'int',
                'pos2': 'int',
                'strand': 'category'
            }
        ).rename_axis('ev_id')

        return df

    @staticmethod
    def _get_donor_acceptor(s):
        chr_, pos1, pos2, strand = s[['chr', 'pos1', 'pos2', 'strand']]
        pos1, pos2 = sorted([pos1, pos2])

        if strand == '+':
            donor_site = pos2
            acceptor_site = pos1
        elif strand == '-':
            donor_site = pos1
            acceptor_site = pos2

        res = [chr_, donor_site, acceptor_site, strand]
        res = pd.Series(res, index=['chr', 'donor', 'acceptor', 'strand'])

        return res

    @classmethod
    def _get_donor_acceptor_df(cls, original_df):
        if original_df.empty:
            df = pd.DataFrame([], columns=['chr', 'donor', 'acceptor', 'strand'])
        else:
            df = original_df.apply(cls._get_donor_acceptor, axis=1)

        return df

    def expand_to_all_events(self, ev_df, fillna_value):
        expanded_df = self.original_df.reset_index().rename(
            {
                'index': 'ev_id'
            },
            axis=1
        ).loc[:, ['ev_id']].merge(
            ev_df,
            on='ev_id',
            how='left'
        ).fillna(fillna_value)

        return expanded_df

    def submit_to_summary(self, summary_column, type_):
        self._summary_columns[type_].append(summary_column)

    def check_annotation(self, anno_db_file):
        self._annotator = Annotator(anno_db_file)
        self.anno_df, anno_status = self.df.pipe(self._annotator.annotate)

        self.submit_to_summary(anno_status, type_='filters')

    def check_ambiguous(self,
                        anno_db_file,
                        ref_file,
                        other_ref_file,
                        work_dir='.',
                        num_proc=1):

        self.checker = AmbiguousChecker(
            anno_db_file,
            ref_file,
            other_ref_file,
            work_dir=work_dir,
            num_proc=num_proc
        )
        checking_result = self.checker.check(self.df)

        self.submit_to_summary(checking_result, type_='filters')

    @staticmethod
    def _merge_columns(df, column_dfs):
        merged_df = df

        for col_df in column_dfs:
            merged_df = merged_df.merge(col_df, on='ev_id', how='left')

        return merged_df

    def get_summary(self):
        filters_df = pd.DataFrame(self.df.index).pipe(
            self._merge_columns,
            self._summary_columns['filters']
        )

        pass_column = filters_df.fillna('2').set_index(
            'ev_id'
        ).agg(
            'sum',
            axis=1
        ).eq(
            0
        ).replace(
            {
                True: 'yes',
                False: 'no'
            }
        ).reset_index().rename({0: 'pass'}, axis=1)

        summary_df = self.original_df.reset_index().pipe(
            self._merge_columns,
            [pass_column] + self._summary_columns['summary'] + [filters_df.fillna('NA')]
        ).set_index('ev_id')

        return summary_df

    @property
    def clear_df(self):
        return self.get_summary()[lambda df: df['pass'] == 'yes'].loc[:, :'strand']

    @property
    def clear_anno_df(self):
        return self.anno_df.merge(
            self.clear_df.reset_index(),
            on='ev_id',
            how='inner'
        )



#! /usr/bin/env python
import click
import os


@click.group()
@click.version_option()
def cli():
    """
    A toolset for investigating the interactions between circRNA, miRNA, and mRNA.
    """

    pass


@cli.command()
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-i', '--circ', 'circ_file', metavar="CIRC_FILE", required=True)
@click.option('-o', '--out-prefix', 'out_prefix', default='./out/', metavar="OUT_PREFIX")
@click.option('-p', '--num_proc', default=1, type=click.INT, metavar="NUM_PROC",
    help="Number of processes")
@click.option('--checkAA', 'checkAA', is_flag=True, help="Check if the circRNA has ambiguous alignments.")
@click.option('--header', 'header', flag_value=True, type=click.BOOL,
              default=True, hidden=True)
@click.option('--no-header', 'header', flag_value=False, type=click.BOOL, hidden=True)
@click.option('--miranda-sc', 'sc', metavar='S', type=click.FLOAT)
@click.option('--miranda-en', 'en', metavar='-E', type=click.FLOAT)
@click.option('--miranda-scale', 'scale', metavar='Z', type=click.FLOAT)
@click.option('--miranda-strict', 'strict', is_flag=True)
@click.option('--miranda-go', 'go', metavar='-X', type=click.FLOAT)
@click.option('--miranda-ge', 'ge', metavar='-Y', type=click.FLOAT)
def run(circ_file, ref_dir, out_prefix, num_proc, header, checkAA, **miranda_options):
    """
    This command is the main pipeline of CircMiMi.
    """

    from circmimi.utils import add_prefix

    output_dir = os.path.dirname(out_prefix)
    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    from circmimi.config import get_refs
    anno_db, ref_file, mir_ref, mir_target, other_transcripts = get_refs(ref_dir)

    summary_file = add_prefix('summary_list.tsv', out_prefix)
    clear_file = add_prefix('circ.clear.tsv', out_prefix)

    if checkAA:
        other_ref_file = other_transcripts
    else:
        other_ref_file = None

    miranda_options_list = []
    for k, v in miranda_options.items():
        if v is not None:
            cmd = ['-' + k]
            if k != 'strict':
                cmd.append(str(v))

            miranda_options_list.extend(cmd)

    from circmimi.circmimi import Circmimi
    circmimi_result = Circmimi(
        anno_db,
        ref_file,
        mir_ref,
        mir_target,
        other_ref_file,
        work_dir=output_dir,
        num_proc=num_proc,
        miranda_options=miranda_options_list
    )

    circmimi_result.run(circ_file)

    result_table = circmimi_result.get_result_table()
    res_file = add_prefix('all_interactions.tsv', out_prefix)
    result_table.to_csv(res_file, sep='\t', index=False, header=header)
    circmimi_result.save_circRNAs_summary(summary_file)
    circmimi_result.save_clear_circRNAs(clear_file)


@cli.command()
@click.option('--species', 'species', metavar="SPECIES_KEY", required=True)
@click.option('--source', 'source', metavar="SOURCE", required=True,
    help="""
        Available sources are "gencode", "ensembl", "ensembl_plants", and "ensembl_metazoa"
    """)
@click.option('--gencode', 'source', flag_value='gencode', type=click.STRING, help="--source gencode", hidden=True)
@click.option('--ensembl', 'source', flag_value='ensembl', type=click.STRING, help="--source ensembl", hidden=True)
@click.option('--version', 'version', default='current', metavar="VERSION",
    help="""
        The release version. If not assigned, it will be automatically set to the latest version of the SOURCE.
    """)
@click.option('--init', 'init', is_flag=True, help="Create an init template ref_dir.", hidden=True)
@click.argument('ref_dir')
def genref(species, source, version, ref_dir, init):
    """
    Generate the references.                                          
    
    The generated reference files will be saved in the REF_DIR.
    """

    os.makedirs(ref_dir, exist_ok=True)

    from circmimi.config import RefConfig
    config = RefConfig()

    if not init:
        from circmimi.reference import genref
        info, ref_files = genref.generate(species, source, version, ref_dir)

        config['info'].update(info)
        config['refs'].update(ref_files)

    config.write(ref_dir)


@cli.command(hidden=True)
@click.argument('gtf_path')
@click.argument('db_path', metavar='OUT_PATH')
def gendb(gtf_path, db_path):
    from circmimi.reference import gendb

    gendb.generate(gtf_path, db_path)


@cli.command(hidden=True)
@click.option('--species', 'species', metavar="SPECIES_KEY", required=True)
@click.option('--version', 'version', default='current', metavar="VERSION", required=True)
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-o', '--out_file', 'out_file', metavar="OUT_FILE", required=True)
@click.option('-a', '--show-accession', 'show_accession', is_flag=True)
def genmirdb(species, version, ref_dir, out_file, show_accession):
    os.makedirs(ref_dir, exist_ok=True)

    from circmimi.reference import genmirdb

    genmirdb.generate(species, version, ref_dir, out_file, show_accession)


@cli.group(hidden=True)
def check():
    pass


@check.command()
@click.argument('circ_file')
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-o', '--out-prefix', 'out_prefix', default='./out/', metavar="OUT_PREFIX")
def annotation(circ_file, ref_dir, out_prefix):
    from circmimi.config import get_refs
    from circmimi.circ import CircEvents
    from circmimi.utils import add_prefix

    output_dir = os.path.dirname(out_prefix)
    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    anno_db, _, _, _, _ = get_refs(ref_dir)

    circ_events = CircEvents(circ_file)
    circ_events.check_annotation(anno_db)
    res_file = add_prefix('circ.summary_list.tsv', out_prefix)
    circ_events.get_summary().to_csv(res_file, sep='\t', index=False)


@check.command()
@click.argument('circ_file')
@click.option('-r', '--ref', 'ref_dir', type=click.Path(), metavar="REF_DIR", required=True)
@click.option('-o', '--out', 'output_dir', metavar="OUT_DIR", required=True)
@click.option('-p', '--num_proc', default=1, type=click.INT, metavar="NUM_PROC",
    help="Number of processes")
def ambiguous(circ_file, ref_dir, output_dir, num_proc):
    os.makedirs(output_dir, exist_ok=True)

    from circmimi.config import get_refs
    anno_db, ref_file, _, _, other_transcripts = get_refs(ref_dir)

    from circmimi.circ import CircEvents
    circ_events = CircEvents(circ_file)
    circ_events.check_ambiguous(
        anno_db,
        ref_file,
        other_transcripts,
        work_dir=output_dir,
        num_proc=num_proc
    )

    result_file = os.path.join(output_dir, 'circ.summary_list.tsv')
    circ_events.get_summary().to_csv(result_file, sep='\t', index=False)


@cli.group()
def network():
    """
    Command to create the network file for Cytoscape.
    """

    pass


@network.command()
@click.argument('in_file')
@click.argument('out_file')
@click.option('-f', '--format', 'format_', default='xgmml',
              help="Assign the format of the OUT_FILE.", hidden=True)
def create(in_file, out_file, format_):
    """
    Create the network file.

    This command would transform the IN_FILE to XGMML format,
    and save to OUT_FILE.
    """

    from circmimi.network.network import CyNetwork, Layout, Style

    network = CyNetwork()
    network.load_data(in_file)

    layout = Layout()
    style = Style()

    network.apply_layout(layout)
    network.apply_style(style)

    if format_ == "xgmml":
        network.to_xgmml(out_file)


@cli.group(hidden=True)
def update_mirna():
    pass


@update_mirna.command()
@click.option('--from', 'from_', required=True)
@click.option('--to', 'to_', required=True)
@click.option('--species')
@click.option('-o', '--out-prefix', 'out_prefix', default='./', metavar="OUT_PREFIX")
def genmaps(from_, to_, species, out_prefix):
    from circmimi.reference.mirbase import MatureMiRNAUpdater
    from circmimi.utils import add_prefix

    output_dir, prefix_name = os.path.split(out_prefix)

    if output_dir == '':
        output_dir = '.'

    if output_dir != '.':
        os.makedirs(output_dir, exist_ok=True)

    updater = MatureMiRNAUpdater(from_, to_, species)
    updater.create(output_dir)

    out_file = 'miRNA.maps_{}_to_{}.tsv'.format(from_, to_)
    out_file = add_prefix(out_file, prefix_name)
    out_file = os.path.join(output_dir, out_file)

    updater.save(out_file)


@update_mirna.command()
@click.argument('in_file')
@click.argument('out_file')
@click.option('-m', '--maps', 'mapping_file', metavar="MAPPING_FILE", required=True)
@click.option('-k', '--col-key', 'column_key', type=click.INT, default=1,
              help="The column number of miRNA IDs.")
@click.option('-i', '--inplace', is_flag=True)
@click.option('-R', '--remove-deleted', is_flag=True)
def update(in_file, out_file, mapping_file, column_key, inplace, remove_deleted):
    column_key = column_key - 1

    from circmimi.reference.mirbase import MatureMiRNAUpdater
    updater = MatureMiRNAUpdater(None, None, None)
    updater.load_maps(mapping_file)
    updater.update_file(in_file, out_file, column_key, inplace, remove_deleted)


if __name__ == "__main__":
    cli()
